"""Entry point to handle queries.

WFS defines 2 query types:
- Adhoc queries are constructed directly from request parameters.
- Stored queries are defined first, and executed later.

Both use the FES (Filter Encoding Syntax) filtering logic internally.

The objects in this module closely follow the WFS spec.
By using the same type definitions, a lot of code logic follows naturally.
The "GetFeatureById" is a mandatory built-in stored query.
"""
from .adhoc import AdhocQuery
from .base import QueryExpression
from .stored import (
    GetFeatureById,
    QueryExpressionText,
    StoredQuery,
    StoredQueryDescription,
    StoredQueryParameter,
    stored_query_registry,
)

__all__ = (
    "QueryExpression",
    "AdhocQuery",
    "QueryExpressionText",
    "StoredQueryDescription",
    "StoredQuery",
    "stored_query_registry",
    "StoredQueryParameter",
    "GetFeatureById",
)



"""Wrappers for the results of GetFeature/GetPropertyValue.

The "SimpleFeatureCollection" and "FeatureCollection" and their
properties match the WFS 2.0 spec closely.
"""
from __future__ import annotations

import math
import operator
from functools import reduce
from typing import Iterable

import django
from django.db import models
from django.utils.functional import cached_property
from django.utils.timezone import now, utc

from gisserver.features import FeatureType
from gisserver.geometries import BoundingBox

from .utils import ChunkedQuerySetIterator, CountingIterator


class SimpleFeatureCollection:
    """Wrapper to read a result set.

    This object type is defined in the WFS spec.
    It holds a collection of "wfs:member" objects.
    """

    def __init__(
        self,
        feature_type: FeatureType,
        queryset: models.QuerySet,
        start: int,
        stop: int,
    ):
        self.feature_type = feature_type
        self.queryset = queryset
        self.start = start
        self.stop = stop
        self._result_cache = None
        self._result_iterator = None

    def __iter__(self) -> Iterable[models.Model]:
        """Iterate through all results.

        Depending on the output format (and whether pagination is read first),
        the results can either be cached first, or be streamed without caching.
        This picks the best-performance scenario in most cases.
        """
        if self.start == self.stop == 0:
            self._result_cache = []

        if self._result_cache is not None:
            # Repeat what's cached if reading again.
            return iter(self._result_cache)
        elif self.queryset._prefetch_related_lookups:
            # Optimal solution to have an iterator that also handles prefetches
            return self._chunked_iterator()
        else:
            # No need to fetch everything first, can just stream the results.
            return self.iterator()

    def iterator(self):
        """Explicitly request the results to be streamed.

        This can be used by output formats that stream may results, and don't
        access `number_returned`. Note this is not compatible with prefetch_related().
        """
        if self._result_iterator is not None:
            raise RuntimeError("Results for feature collection are read twice.")

        if self._result_cache is not None:
            # In case the results were read already, reuse that.
            return iter(self._result_cache)
        elif self.start == self.stop == 0:
            # resulttype=hits
            return iter([])
        else:
            model_iter = self._paginated_queryset().iterator()
            self._result_iterator = CountingIterator(model_iter)
            return iter(self._result_iterator)

    def _chunked_iterator(self):
        """Generate an interator that processes results in chunks."""
        # Private function so the same logic of .iterator() is not repeated.
        self._result_iterator = ChunkedQuerySetIterator(self._paginated_queryset())
        return iter(self._result_iterator)

    def _paginated_queryset(self) -> models.QuerySet:
        """Apply the pagination to the queryset."""
        if self.stop == math.inf:
            # Infinite page requested
            if self.start:
                return self.queryset[self.start :]
            else:
                return self.queryset
        else:
            return self.queryset[self.start : self.stop]

    def first(self):
        try:
            # Don't query a full page, return only one instance (for GetFeatureById)
            # This also preserves the extra added annotations (like _as_gml_FIELD)
            return self.queryset[self.start]
        except IndexError:
            return None

    def fetch_results(self):
        """Forcefully read the results early."""
        if self._result_cache is not None:
            return
        if self._result_iterator is not None:
            raise RuntimeError("Results for feature collection are read twice.")

        if self.start == self.stop == 0:
            # resulttype=hits
            self._result_cache = []
        else:
            # This still allows prefetch_related() to work,
            # since QuerySet.iterator() is avoided.
            if self.stop == math.inf:
                # Infinite page requested
                if self.start:
                    qs = self.queryset[self.start :]
                else:
                    qs = self.queryset.all()
            else:
                qs = self.queryset[self.start : self.stop]

            self._result_cache = list(qs)

    @cached_property
    def number_returned(self) -> int:
        """Return the number of results for this page."""
        if self.start == self.stop == 0:
            return 0  # resulttype=hits
        elif self._result_iterator is not None:
            # When requesting the data after the fact, results are counted.
            return self._result_iterator.number_returned
        else:
            # Count by fetching all data. Otherwise the results are queried twice.
            # For GML/XML, it's not possible the stream the queryset results
            # as the first tag needs to describe the number of results.
            self.fetch_results()
            return len(self._result_cache)

    @cached_property
    def number_matched(self) -> int:
        """Return the total number of matches across all pages."""
        page_size = self.stop - self.start
        if page_size and self.number_returned < page_size:
            # For resulttype=results, an expensive COUNT query can be avoided
            # when this is the first and only page or the last page.
            return self.start + self.number_returned

        qs = self.queryset
        clean_annotations = {
            # HACK: remove database optimizations from output renderer.
            # Otherwise it becomes SELECT COUNT(*) FROM (SELECT AsGML(..), ...)
            key: value
            for key, value in qs.query.annotations.items()
            if not key.startswith("_as_")
        }
        if clean_annotations != qs.query.annotations:
            qs = self.queryset.all()  # make a clone to allow editing
            if django.VERSION >= (3, 0):
                qs.query.annotations = clean_annotations
            else:
                qs.query._annotations = clean_annotations

        return qs.count()

    def get_bounding_box(self) -> BoundingBox:
        """Determine bounding box of all items."""
        self.fetch_results()  # Avoid querying results twice

        # Start with an obviously invalid bbox,
        # which corrects at the first extend_to_geometry call.
        bbox = BoundingBox(math.inf, math.inf, -math.inf, -math.inf)
        geometry_field = self.feature_type.resolve_element(
            self.feature_type.geometry_field.name
        ).child
        for instance in self:
            geomery_value = geometry_field.get_value(instance)
            if geomery_value is None:
                continue

            bbox.extend_to_geometry(geomery_value)

        return bbox


CALCULATE = -9999999


class FeatureCollection:
    """WFS object that holds the result type for GetFeature.
    This object type is defined in the WFS spec.
    """

    def __init__(
        self,
        results: list[SimpleFeatureCollection],
        number_matched: int | None = CALCULATE,
        next: str | None = None,
        previous: str | None = None,
    ):
        """
        :param results: All retrieved feature collections (one per FeatureType)
        :param number_matched: Total number of features across all pages
        :param next: URL of the next page
        :param previous: URL of the previous page
        """
        self.results = results
        self._number_matched = number_matched
        self.next = next
        self.previous = previous
        self.date = now()
        self.timestamp = self.date.astimezone(utc).isoformat()

    def get_bounding_box(self) -> BoundingBox:
        """Determine bounding box of all items."""
        # Combine the bounding box of all collections
        return reduce(operator.add, [c.get_bounding_box() for c in self.results])

    @cached_property
    def number_returned(self) -> int:
        """Return the total number of returned features"""
        return sum(c.number_returned for c in self.results)

    @cached_property
    def number_matched(self) -> int | None:
        """The number of features matched, None means "unknown"."""
        if self._number_matched is None:
            return None  # WFS allows returning this as "unknown"
        elif self._number_matched == CALCULATE:
            # By making the number_matched lazy, the calling method has a chance to
            # decorate the results with extra annotations before they are evaluated.
            # This is needed since SimpleCollection.number_matched already evaluates the queryset.
            return sum(c.number_matched for c in self.results)
        else:
            # Evaluate any lazy attributes
            return int(self._number_matched)

    def __iter__(self):
        return iter(self.results)



"""The view layer parses the request, and dispatches it to an operation."""
from __future__ import annotations

import re
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import PermissionDenied as Django_PermissionDenied
from django.core.exceptions import SuspiciousOperation
from django.http import HttpResponse
from django.shortcuts import render
from django.views import View

from gisserver import conf
from gisserver.exceptions import (
    InvalidParameterValue,
    MissingParameterValue,
    OperationNotSupported,
    OWSException,
    PermissionDenied,
)
from gisserver.features import FeatureType, ServiceDescription
from gisserver.operations import base, wfs20

SAFE_VERSION = re.compile(r"\A[0-9.]+\Z")


class GISView(View):
    """The base logic to implement OGC view like WFS.

    Each subclass defines 'accept_operations' with the desired RPC operations.
    """

    #: Define the namespace to use in the XML
    xml_namespace = "http://example.org/gisserver"

    #: Default version to use
    version = "2.0.0"

    #: Allow to set a default service, so the SERVICE parameter can be omitted.
    default_service = None

    #: Supported versions
    accept_versions = ("2.0.0",)

    #: Internal configuration of all available RPC calls.
    accept_operations = {}

    #: Metadata of the service:
    service_description: ServiceDescription | None = None

    #: Template to render a HTML welcome page for non-OGC requests.
    index_template_name = None

    #: Whether to render GET HTML pages
    use_html_templates = True

    def dispatch(self, request, *args, **kwargs):
        """Render proper XML errors for exceptions on all request types."""
        try:
            return super().dispatch(request, *args, **kwargs)
        except Exception as e:
            response = self.handle_exception(e)
            if response is not None:
                return response
            raise

    def handle_exception(self, exc):
        """Transform an exception into a OGC response.
        When nothing is returned, the exception is raised instead.
        """
        if isinstance(exc, Django_PermissionDenied):
            exc = PermissionDenied("typeNames", text=str(exc) or None)
            return HttpResponse(
                exc.as_xml().encode("utf-8"),
                content_type="text/xml; charset=utf-8",
                status=exc.status_code,
                reason=exc.reason,
            )
        elif isinstance(exc, OWSException):
            # Wrap our XML-based exception into a response.
            exc.version = self.version  # Use negotiated version
            return HttpResponse(
                exc.as_xml().encode("utf-8"),
                content_type="text/xml; charset=utf-8",
                status=exc.status_code,
                reason=exc.reason,
            )

    def get(self, request, *args, **kwargs):
        """Entry point to handle HTTP GET requests.

        This parses the 'SERVICE' and 'REQUEST' parameters,
        to call the proper operation.

        All query parameters are handled as case-insensitive.
        """
        # Convert to WFS key-value-pair format.
        self.KVP = {key.upper(): value for key, value in request.GET.items()}

        # Perform early version parsing. The detailed validation happens by the operation
        # Parameter objects, but by performing an early check, templates can use that version.
        version = self.KVP.get("VERSION")
        if version and version in self.accept_versions:
            self.set_version(version)

        # Allow for an user-friendly opening page (hence the version check above)
        if self.use_html_templates and self.is_index_request():
            return self.render_index()

        # Normal WFS
        wfs_method_cls = self.get_operation_class()
        return self.call_operation(wfs_method_cls)

    def is_index_request(self):
        """Tell whether to index page should be shown."""
        # If none of the typical WFS parameters are given, and there aren't many other
        # (misspelled?) parameters on the query string, this is considered to be an index page.
        # Minimal request has 2 parameters (SERVICE=WFS&REQUEST=GetCapabilities).
        # Some servers also allow a default of SERVICE, thus needing even less.
        return len(self.KVP) < 2 and {
            "REQUEST",
            "SERVICE",
            "VERSION",
            "ACCEPTVERSIONS",
        }.isdisjoint(self.KVP.keys())

    def render_index(self):
        """Render the index page."""
        # Not using the whole TemplateResponseMixin config with configurable parameters.
        # If this really needs more configuration, overriding is likely just as easy.
        return render(
            self.request, self.get_index_template_names(), self.get_index_context_data()
        )

    def get_index_context_data(self, **kwargs):
        """Provide the context data for the index page."""
        service = self.KVP.get("SERVICE", self.default_service)
        root_url = self.request.build_absolute_uri()

        # Allow passing extra vendor parameters to the links generated in the template
        # (e.g. expand/embed logic)
        base_qs = {
            key: value
            for key, value in self.request.GET.items()
            if key.upper()
            not in ("SERVICE", "REQUEST", "VERSION", "OUTPUTFORMAT", "TYPENAMES")
        }
        base_query = urlencode(base_qs) + "&" if base_qs else ""

        return {
            "view": self,
            "service": service,
            "root_url": root_url,
            "base_query": base_query,
            "connect_url": root_url,
            "version": self.version,
            "service_description": self.get_service_description(service),
            "accept_versions": self.accept_versions,
            "accept_operations": self.accept_operations,
            **kwargs,
        }

    def get_index_template_names(self):
        """Get the index page template name.
        If no template is configured, some reasonable defaults are selected.
        """
        raw_service = self.KVP.get("SERVICE", self.default_service)
        if raw_service and raw_service in self.accept_operations:
            service = raw_service.lower()
        else:
            service = "default"

        if self.index_template_name:
            # Allow the same substitutions for a manually configured template.
            return [
                self.index_template_name.format(service=service, version=self.version)
            ]
        else:
            return [
                f"gisserver/{service}/{self.version}/index.html",
                f"gisserver/{service}/index.html",
                "gisserver/index.html",
            ]

    def get_operation_class(self) -> type[base.WFSMethod]:
        """Resolve the method that the client wants to call."""
        if not self.accept_operations:
            raise ImproperlyConfigured("View has no operations")

        # The service is always WFS
        service = self._get_required_arg("SERVICE", self.default_service).upper()
        try:
            operations = self.accept_operations[service]
        except KeyError:
            allowed = ", ".join(sorted(self.accept_operations.keys()))
            raise InvalidParameterValue(
                "service",
                f"'{service}' is an invalid service, supported are: {allowed}.",
            ) from None

        # Resolve the operation
        # In mapserver, the operation name is case insensitive.
        operation = self._get_required_arg("REQUEST").upper()
        uc_methods = {name.upper(): method for name, method in operations.items()}

        try:
            return uc_methods[operation]
        except KeyError:
            allowed = ", ".join(operations.keys())
            raise OperationNotSupported(
                "request",
                f"'{operation.lower()}' is not implemented, supported are: {allowed}.",
            ) from None

    def call_operation(self, wfs_method_cls: type[base.WFSMethod]):
        """Call the resolved method."""
        wfs_method = wfs_method_cls(self)
        param_values = wfs_method.parse_request(self.KVP)
        return wfs_method(**param_values)  # goes into __call__()

    def _get_required_arg(self, argname, default=None):
        try:
            return self.KVP[argname]
        except KeyError:
            if default is not None:
                return default
            raise MissingParameterValue(argname.lower()) from None

    def get_service_description(self, service: str) -> ServiceDescription:
        """Provide the (dynamically generated) service description."""
        return self.service_description or ServiceDescription(title="Unnamed")

    def set_version(self, version):
        """Enforce a particular version based on the request."""
        if not SAFE_VERSION.match(version):
            # Make really sure we didn't mess things up, as this causes file includes.
            raise SuspiciousOperation("Invalid/insecure version number parsed")

        # Enforce the requested version
        self.version = version

    @property
    def server_url(self):
        """Expose the server URLs for all operations to read."""
        return self.request.build_absolute_uri(self.request.path)


class WFSView(GISView):
    """A view for a single WFS server.

    This view exposes multiple dataset,
    each containing a single Django model (mapped to feature types in WFS).

    This class can be used by subclassing it, and redefining ``feature_types``
    or ``get_feature_types()``.
    """

    #: Maximum number of features to return
    max_page_size = conf.GISSERVER_DEFAULT_MAX_PAGE_SIZE

    #: Define the features (=tables) in this dataset.
    feature_types: list[FeatureType] = []

    #: Internal configuration of all available RPC calls.
    accept_operations = {
        "WFS": {
            "GetCapabilities": wfs20.GetCapabilities,
            "DescribeFeatureType": wfs20.DescribeFeatureType,
            "GetFeature": wfs20.GetFeature,
            "GetPropertyValue": wfs20.GetPropertyValue,
            "ListStoredQueries": wfs20.ListStoredQueries,
            "DescribeStoredQueries": wfs20.DescribeStoredQueries,
        }
    }

    #: Since URLs to this view are already specifically for WFS,
    #: allow to omit the service name.
    default_service = "WFS"

    #: Metadata of the capabilities:
    wfs_service_constraints = {
        "ImplementsBasicWFS": True,  # Advanced queries/xpath
        "ImplementsTransactionalWFS": False,  # only reads
        "ImplementsLockingWFS": False,  # only basic WFS
        "KVPEncoding": True,  # HTTP GET support
        "XMLEncoding": False,  # HTTP POST requests
        "SOAPEncoding": False,  # no SOAP requests
        "ImplementsInheritance": False,
        "ImplementsRemoteResolve": False,
        "ImplementsResultPaging": True,  # missing next/previous for START / COUNT
        "ImplementsStandardJoins": False,  # returns records as wfs:Tuple in GetFeature
        "ImplementsSpatialJoins": False,
        "ImplementsTemporalJoins": False,
        "ImplementsFeatureVersioning": False,
        "ManageStoredQueries": False,
        #
        # Mentioned as operation constraint in WFS spec:
        "PagingIsTransactionSafe": False,
    }

    #: Metadata of filtering capabilities
    wfs_filter_capabilities = {
        "ImplementsQuery": True,  # <fes:AbstractQueryElement> needed for WFS simple
        "ImplementsAdHocQuery": True,  # <fes:AbstractAdhocQueryElement> needed for WFS simple
        "ImplementsFunctions": True,  # <fes:Function> support
        "ImplementsResourceId": True,  # <fes:ResourceId> support
        "ImplementsMinStandardFilter": True,  # <fes:PropertyIs...> support
        "ImplementsStandardFilter": True,  # <fes:And>, <fes:Or> and advanced functions
        "ImplementsMinSpatialFilter": True,  # <fes:BBOX>
        "ImplementsSpatialFilter": True,  # Other spatial functions
        "ImplementsMinTemporalFilter": False,  # mapserver: true (During)
        "ImplementsTemporalFilter": False,
        "ImplementsVersionNav": False,  # <fes:ResourceId version="..">
        "ImplementsSorting": True,  # SORTBY parameter
        "ImplementsExtendedOperators": False,
        "ImplementsMinimumXPath": True,  # incomplete like mapserver, needed for cite
        "ImplementsSchemaElementFunc": False,
    }

    def get_feature_types(self) -> list[FeatureType]:
        """Return all available feature types this server exposes"""
        return self.feature_types

    def get_index_context_data(self, **kwargs):
        """Add WFS specific metadata"""
        wfs_output_formats = self.accept_operations["WFS"]["GetFeature"].output_formats

        context = super().get_index_context_data(**kwargs)
        context.update(
            {
                "wfs_features": self.get_feature_types(),
                "wfs_output_formats": wfs_output_formats,
                "wfs_filter_capabilities": self.wfs_filter_capabilities,
                "wfs_service_constraints": self.wfs_service_constraints,
            }
        )
        return context



"""Handle (stored)query objects.

These definitions follow the WFS spec.
"""
from __future__ import annotations

from dataclasses import dataclass

from django.db.models import Q, QuerySet

from gisserver.exceptions import InvalidParameterValue, MissingParameterValue, NotFound
from gisserver.features import FeatureType
from gisserver.operations.base import Parameter
from gisserver.output import FeatureCollection
from gisserver.parsers import fes20
from gisserver.types import XsdTypes

from .base import QueryExpression


class QueryExpressionText:
    """Define the body of a stored query.

    This object type is defined in the WFS spec.
    It may contain a wfs:Query or wfs:StoredQuery element.
    """

    return_feature_types: list[str] | None = None
    language: str = fes20.Filter.query_language
    is_private: bool = False


@dataclass
class StoredQueryDescription:
    """WFS metadata of a stored query.
    This object type is defined in the WFS spec.
    """

    id: str = None
    title: str | None = None
    abstract: str | None = None
    parameters: dict[str, XsdTypes] | None = None
    expressions: list = None  # TODO: support multiple body expressions


class StoredQuery(QueryExpression):
    """Base class for stored queries.

    This represents all predefined queries on the server.
    A good description can be found at:
    https://mapserver.org/ogc/wfs_server.html#stored-queries-wfs-2-0

    The implementation logic is fully defined by the :class:`QueryExpression`
    base class. For a simple implementation, the following data should be
    overwritten:

    * :meth:`get_type_names` to define which type this query references.
    * :meth:`compile_query` to define the queryset filter.

    For advanced overriding, see the :class:`QueryExpression` base class,
    or the :class:`GetFeatureById` implementation.
    """

    # Official WFS docs have an 'id' and 'parameters' property for the
    # StoredQuery class, but these are avoided here to give subclasses full
    # control over which properties to store. E.g. "id" conflicts with Django
    # model subclasses that stores the query.
    meta: StoredQueryDescription

    def __init__(self, **parameters):
        self.parameters = parameters

    @classmethod
    def extract_parameters(cls, KVP) -> dict[str, str]:
        """Extract the arguments from the key-value-pair (=HTTP GET) request."""
        args = {}
        for name, _xsd_type in cls.meta.parameters.items():
            try:
                args[name] = KVP[name]
            except KeyError:
                raise MissingParameterValue(
                    name, f"Stored query {cls.meta.id} requires an '{name}' parameter"
                ) from None

        # Avoid unexpected behavior, check whether the client also sends adhoc query parameters
        for name in ("filter", "bbox", "resourceID"):
            if name not in args and KVP.get(name.upper()):
                raise InvalidParameterValue(
                    name, "Stored query can't be combined with adhoc-query parameters"
                )

        return args


class StoredQueryRegistry:
    """Registry of functions to be callable by <fes:Query>."""

    def __init__(self):
        self.stored_queries = {}

    def __bool__(self):
        return bool(self.stored_queries)

    def __iter__(self):
        return iter(self.stored_queries.values())

    def register(self, meta: StoredQueryDescription | None = None, **meta_kwargs):
        """Register a custom class that handles a stored query"""

        def _metadata_dec(query: type[StoredQuery]):
            query.meta = meta or StoredQueryDescription(**meta_kwargs)
            self.stored_queries[query.meta.id] = query
            return query

        return _metadata_dec

    def resolve_query(self, query_id) -> type[StoredQuery]:
        """Find the stored procedure using the ID."""
        try:
            return self.stored_queries[query_id]
        except KeyError:
            raise InvalidParameterValue(
                "STOREDQUERY_ID", f"Stored query does not exist: {query_id}"
            ) from None


stored_query_registry = StoredQueryRegistry()


class StoredQueryParameter(Parameter):
    """Special parameter parsing for the 'STOREDQUERY_ID' parameter"""

    def __init__(self):
        super().__init__(
            name="STOREDQUERY_ID", parser=stored_query_registry.resolve_query
        )

    def value_from_query(self, KVP: dict):
        """Customize the request parsing to read custom parameters too."""
        stored_query_class = super().value_from_query(KVP)
        if stored_query_class is None:
            return None

        parameters = stored_query_class.extract_parameters(KVP)
        return stored_query_class(**parameters)


@stored_query_registry.register(
    id="urn:ogc:def:query:OGC-WFS::GetFeatureById",
    title="Get feature by identifier",
    abstract="Returns the single feature that corresponds with the ID argument",
    parameters={"ID": XsdTypes.string},
    # expressions=[QueryExpressionText],
)
class GetFeatureById(StoredQuery):
    """The stored query for GetFeatureById.

    This is essentially the same as:

    <wfs:Query xmlns:wfs='..." xmlns:fes='...'>
        <fes:Filter><fes:ResourceId rid='{ID}'/></fes:Filter>
    </wfs:Query>

    Except that the response is supposed to contain only the item itself.
    """

    def __init__(self, ID):
        super().__init__(ID=ID)
        try:
            type_name, id = ID.rsplit(".", 1)
        except ValueError:
            # Always report this as 404
            raise NotFound("ID", "Expected typeName.id for ID parameter") from None

        self.type_name = type_name
        self.id = id

    def get_type_names(self) -> list[FeatureType]:
        """Tell which type names this query applies to."""
        feature_type = self.all_feature_types[self.type_name]
        return [feature_type]

    def get_queryset(self, feature_type: FeatureType) -> QuerySet:
        """Override to implement ID type checking."""
        try:
            return super().get_queryset(feature_type)
        except (ValueError, TypeError) as e:
            raise InvalidParameterValue("ID", f"Invalid ID value: {e}") from e

    def get_results(self, *args, **kwargs) -> FeatureCollection:
        """Override to implement 404 checking."""
        collection = super().get_results(*args, **kwargs)

        # Directly attempt to collect the data.
        # Avoid having to do that in the output renderer.
        if collection.results[0].first() is None:
            # WFS 2.0.2: Return NotFound instead of InvalidParameterValue
            raise NotFound("ID", f"Feature not found with ID {self.id}.")

        return collection

    def compile_query(
        self, feature_type: FeatureType, using=None
    ) -> fes20.CompiledQuery:
        """Create the internal query object that will be applied to the queryset."""
        compiler = fes20.CompiledQuery(feature_type=feature_type)
        compiler.add_lookups(Q(pk=self.id), type_name=self.type_name)
        return compiler



from __future__ import annotations

from django.db.models import QuerySet

from gisserver.exceptions import InvalidParameterValue
from gisserver.features import FeatureType
from gisserver.output import FeatureCollection, SimpleFeatureCollection
from gisserver.parsers import fes20
from gisserver.types import split_xml_name


class QueryExpression:
    """WFS base class for all queries.
    This object type is defined in the WFS spec (as <fes:AbstractQueryExpression>).

    The subclasses can override the following logic:

    * :meth:`get_type_names` defines which types this query applies to.
    * :meth:`compile_query` defines how to filter the queryset.

    For full control, these methods can also be overwritten instead:

    * :meth:`get_queryset` defines the full results.
    * :meth:`get_hits` to return the collection for RESULTTYPE=hits.
    * :meth:`get_results` to return the collection for RESULTTYPE=results
    """

    handle = ""
    value_reference = None

    def bind(
        self,
        all_feature_types: dict[str, FeatureType],
        value_reference: fes20.ValueReference | None,
    ):
        """Bind the query to presentation-layer logic"""
        self.all_feature_types = all_feature_types
        self.value_reference = value_reference

    def check_permissions(self, request):
        """Verify whether the user has access to view these data sources"""
        for feature_type in self.get_type_names():
            feature_type.check_permissions(request)

    def resolve_type_name(self, type_name, locator="typename") -> FeatureType:
        """Find the feature type for a given name.
        This is an utility that cusstom subclasses can use.
        """
        # Strip the namespace prefix. The Python ElementTree parser does
        # not expose the used namespace prefixes, so text-values can't be
        # mapped against it. As we expose just one namespace, just strip it.
        xmlns, type_name = split_xml_name(type_name)

        try:
            return self.all_feature_types[type_name]
        except KeyError:
            raise InvalidParameterValue(
                locator, f"Typename '{type_name}' doesn't exist in this server."
            ) from None

    def get_hits(self) -> FeatureCollection:
        """Run the query, return the number of hits only.

        Override this method in case you need full control over the response data.
        Otherwise, override :meth:`compile_query` or :meth:`get_queryset`.
        """
        querysets = self.get_querysets()
        return FeatureCollection(
            results=[
                # Include empty feature collections,
                # so the selected feature types are still known.
                SimpleFeatureCollection(
                    feature_type=ft, queryset=qs.none(), start=0, stop=0
                )
                for ft, qs in querysets
            ],
            number_matched=sum(qs.count() for ft, qs in querysets),
        )

    def get_results(self, start_index=0, count=100) -> FeatureCollection:
        """Run the query, return the full paginated results.

        Override this method in case you need full control over the response data.
        Otherwise, override :meth:`compile_query` or :meth:`get_queryset`.
        """
        stop = start_index + count

        # The querysets are not executed yet, until the output is reading them.
        querysets = self.get_querysets()
        return FeatureCollection(
            results=[
                SimpleFeatureCollection(feature_type, qs, start=start_index, stop=stop)
                for feature_type, qs in querysets
            ]
        )

    def get_querysets(self) -> list[tuple[FeatureType, QuerySet]]:
        """Construct the querysets that return the database results."""
        results = []
        for feature_type in self.get_type_names():
            queryset = self.get_queryset(feature_type)
            results.append((feature_type, queryset))

        return results

    def get_queryset(self, feature_type: FeatureType) -> QuerySet:
        """Generate the queryset for the specific feature type.

        This method can be overwritten in subclasses to define the returned data.
        However, consider overwriting :meth:`compile_query` instead of simple data.
        """
        queryset = feature_type.get_queryset()

        # Apply filters
        compiler = self.compile_query(feature_type, using=queryset.db)

        if self.value_reference is not None:
            if feature_type.resolve_element(self.value_reference.xpath) is None:
                raise InvalidParameterValue(
                    "valueReference",
                    f"Field '{self.value_reference.xpath}' does not exist.",
                )

            # For GetPropertyValue, adjust the query so only that value is requested.
            # This makes sure XPath attribute selectors are already handled by the
            # database query, instead of being a presentation-layer handling.
            field = compiler.add_value_reference(self.value_reference)
            queryset = compiler.filter_queryset(queryset, feature_type=feature_type)
            return queryset.values("pk", member=field)
        else:
            return compiler.filter_queryset(queryset, feature_type=feature_type)

    def get_type_names(self) -> list[FeatureType]:
        """Tell which type names this query applies to.

        This method needs to be defined in subclasses.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}.get_type_names() should be implemented."
        )

    def compile_query(
        self, feature_type: FeatureType, using=None
    ) -> fes20.CompiledQuery:
        """Define the compiled query that filters the queryset.

        Subclasses need to define this method, unless
        :meth:`get_queryset` is completely overwritten.
        """
        raise NotImplementedError()



"""Generic support for various GML versions.

These functions locate GML objects, and redirect to the proper parser.
"""
from __future__ import annotations

from xml.etree.ElementTree import Element

from defusedxml.ElementTree import fromstring

from gisserver.exceptions import ExternalParsingError
from gisserver.parsers.base import tag_registry

from .base import AbstractGeometry, GM_Envelope, GM_Object, TM_Object
from .geometries import is_gml_element  # also do tag registration

# All known root nodes as GML object:
FES_GML_NODES = (GM_Object, GM_Envelope, TM_Object)

__all__ = [
    "GM_Object",
    "GM_Envelope",
    "TM_Object",
    "AbstractGeometry",
    "parse_gml",
    "parse_gml_node",
    "find_gml_nodes",
]


def parse_gml(text: str | bytes) -> FES_GML_NODES:
    """Parse an XML <gml:...> string."""
    root_element = fromstring(text)
    return parse_gml_node(root_element)


def parse_gml_node(element: Element) -> FES_GML_NODES:
    """Parse the element"""
    if not is_gml_element(element):
        raise ExternalParsingError(f"Expected GML namespace for {element.tag}")

    return tag_registry.from_child_xml(element, allowed_types=FES_GML_NODES)


def find_gml_nodes(element: Element) -> list[Element]:
    """Find all gml elements in a node"""
    result = []
    for child in element:
        # This selects all GML elements, including 2.1
        if is_gml_element(child):
            result.append(child)

    return result



"""Exceptions for the WFS operations.

Not all exception codes are listed here, only the one that
apply to the current conformance class.

See:
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#35
https://docs.opengeospatial.org/is/09-025r2/09-025r2.html#411
"""
from django.conf import settings
from django.utils.html import format_html


class ExternalValueError(ValueError):
    """Raise a ValueError for external input.
    This helps to distinguish between internal bugs
    (e.g. unpacking values) and misformed external input.
    """


class ExternalParsingError(ValueError):
    """Raise a ValueError for a parsing problem."""


class OWSException(Exception):
    """Base class for XML based exceptions in this module."""

    status_code = 400  # Most common code in spec
    reason = None
    service = None
    version = "2.0.0"
    code = None
    text_template = None

    def __init__(self, locator, text=None, code=None, status_code=None):
        text = text or self.text_template.format(code=self.code, locator=locator)
        super().__init__(text)
        self.locator = locator
        self.text = text
        self.code = code or self.code
        self.status_code = status_code or self.status_code

    def as_xml(self):
        return format_html(
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            "<ows:ExceptionReport"
            ' xmlns:ows="http://www.opengis.net/ows/1.1"'
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
            ' xsi:schemaLocation="http://www.opengis.net/ows/1.1'
            ' http://schemas.opengis.net/ows/1.1.0/owsExceptionReport.xsd"'
            ' xml:lang="en-US"'
            ' version="2.0.0">\n'
            '  <ows:Exception exceptionCode="{code}" locator="{locator}">\n'
            "    <ows:ExceptionText>{text}{debug}</ows:ExceptionText>\n"
            "  </ows:Exception>\n"
            "</ows:ExceptionReport>",
            code=self.code,
            locator=self.locator,
            text=self.text,
            debug=(
                ".\n\n(set GISSERVER_WRAP_FILTER_DB_ERRORS=False to see the Django error page)"
                if settings.DEBUG and self.status_code >= 500
                else ""
            ),
        )

    def __html__(self):
        return self.as_xml()


class WFSException(OWSException):
    service = "WFS"


class OperationNotSupported(WFSException):
    """WFS Method is called, but does not exist on this server."""

    status_code = 400
    reason = "Not Implemented"
    code = "OperationNotSupported"
    text_template = "Operation is not implemented."


class OperationParsingFailed(WFSException):
    """Error parsing the request to a WFS method."""

    status_code = 400
    code = "OperationParsingFailed"
    text_template = "The request could not be parsed by the server."


class OperationProcessingFailed(WFSException):
    """Error while processing the request."""

    status_code = 500  # 400 & 403 may also be used but are deprecated.
    reason = "Server processing failed"
    code = "OperationProcessingFailed"
    text_template = "The request could not be processed by the server."


class MissingParameterValue(WFSException):
    """Required parameter is missing"""

    status_code = 400
    code = "MissingParameterValue"
    text_template = "Missing required '{locator}' parameter."


class InvalidParameterValue(WFSException):
    """Unsupported choice, e.g. unsupported CRS, bad resultType."""

    status_code = 400
    code = "InvalidParameterValue"
    text_template = "Invalid value for '{locator}' parameter."


class VersionNegotiationFailed(WFSException):
    """GetCapabilities called with unsupported versions."""

    status_code = 400
    code = "VersionNegotiationFailed"
    text_template = "'ACCEPTVERSIONS' contains an invalid version number."


class NotFound(WFSException):
    """The requested ResourceId could not be found."""

    status_code = 404
    reason = "Invalid feature or property value"
    code = "NotFound"
    text_template = "Invalid feature or property value"


class OptionNotSupported(WFSException):
    status_code = 400
    reason = "Not Implemented"
    code = "OptionNotSupported"
    text_template = "Option is not implemented."


class PermissionDenied(WFSException):
    """Permission denied (custom error).

    Note this error is not part of the spec,
    but it's still useful to have access controls.
    """

    status_code = 403
    reason = "Permission Denied"
    code = "PermissionDenied"
    text_template = "You do not have permission to perform this action."



"""All supported output formats"""
from gisserver import conf

from .base import OutputRenderer
from .csv import CSVRenderer, DBCSVRenderer
from .geojson import DBGeoJsonRenderer, GeoJsonRenderer
from .gml32 import (
    DBGML32Renderer,
    DBGML32ValueRenderer,
    GML32Renderer,
    GML32ValueRenderer,
)
from .results import FeatureCollection, SimpleFeatureCollection
from .xmlschema import XMLSchemaRenderer

try:
    from django.utils.functional import classproperty  # Django 3.1+
except ImportError:
    from django.utils.decorators import classproperty

__all__ = [
    "OutputRenderer",
    "FeatureCollection",
    "SimpleFeatureCollection",
    "DBCSVRenderer",
    "DBGeoJsonRenderer",
    "DBGML32Renderer",
    "DBGML32ValueRenderer",
    "CSVRenderer",
    "GeoJsonRenderer",
    "GML32Renderer",
    "GML32ValueRenderer",
    "XMLSchemaRenderer",
    "geojson_renderer",
    "gml32_renderer",
    "gml32_value_renderer",
]


def select_renderer(native_renderer_class, db_renderer_class):
    """Dynamically select the preferred renderer.
    This allows changing the settings within the app.
    """

    class SelectRenderer:
        @classproperty
        def real_class(self):
            if conf.GISSERVER_USE_DB_RENDERING:
                return db_renderer_class
            else:
                return native_renderer_class

        def __new__(cls, *args, **kwargs):
            # Return the actual class instead
            return cls.real_class(*args, **kwargs)

        @classmethod
        def decorate_collection(cls, *args, **kwargs):
            return cls.real_class.decorate_collection(*args, **kwargs)

        @classproperty
        def max_page_size(cls):
            return cls.real_class.max_page_size

    return SelectRenderer


csv_renderer = select_renderer(CSVRenderer, DBCSVRenderer)
gml32_renderer = select_renderer(GML32Renderer, DBGML32Renderer)
gml32_value_renderer = select_renderer(GML32ValueRenderer, DBGML32ValueRenderer)
geojson_renderer = select_renderer(GeoJsonRenderer, DBGeoJsonRenderer)



from __future__ import annotations

from collections import defaultdict
from itertools import islice
from typing import Iterable, TypeVar

from django.db import models
from lru import LRU

M = TypeVar("M", bound=models.Model)

DEFAULT_SQL_CHUNK_SIZE = 2000  # allow unit tests to alter this.


class CountingIterator(Iterable[M]):
    """A simple iterator that counts how many results are given."""

    def __init__(self, iterator: Iterable[M]):
        self._iterator = iterator
        self._number_returned = 0
        self._in_iterator = False

    def __iter__(self):
        # Count the number of returned items while reading them.
        # Tried using map(itemgetter(0), zip(model_iter, count_iter)) but that's not faster.
        self._in_iterator = True
        try:
            self._number_returned = 0
            for instance in self._iterator:
                self._number_returned += 1
                yield instance
        finally:
            self._in_iterator = False

    @property
    def number_returned(self) -> int:
        """Tell how many objects the iterator processed"""
        if self._in_iterator:
            raise RuntimeError("Can't read number of returned results during iteration")
        return self._number_returned


class ChunkedQuerySetIterator(Iterable[M]):
    """An optimal strategy to perform ``prefetch_related()`` on large datasets.

    It fetches data from the queryset in chunks,
    and performs ``prefetch_related()`` behavior on each chunk.

    Django's ``QuerySet.prefetch_related()`` works by loading the whole queryset into memory,
    and performing an analysis of the related objects to fetch. When working on large datasets,
    this is very inefficient as more memory is consumed. Instead, ``QuerySet.iterator()``
    is preferred here as it returns instances while reading them. Nothing is stored in memory.
    Hence, both approaches are fundamentally incompatible. This class performs a
    mixed strategy: load a chunk, and perform prefetches for that particular batch.

    As extra performance benefit, a local cache avoids prefetching the same records
    again when the next chunk is analysed. It has a "least recently used" cache to avoid
    flooding the caches when foreign keys constantly point to different unique objects.
    """

    def __init__(self, queryset: models.QuerySet, chunk_size=None, sql_chunk_size=None):
        """
        :param queryset: The queryset to iterate over, that has ``prefetch_related()`` data.
        :param chunk_size: The size of each segment to analyse in-memory for related objects.
        :param sql_chunk_size: The size of each segment to fetch from the database,
            used when server-side cursors are not available. The default follows Django behavior.
        """
        self.queryset = queryset
        self.sql_chunk_size = sql_chunk_size or DEFAULT_SQL_CHUNK_SIZE
        self.chunk_size = chunk_size or self.sql_chunk_size
        self._fk_caches = defaultdict(lambda: LRU(self.chunk_size // 2))
        self._number_returned = 0
        self._in_iterator = False

    def __iter__(self):
        # Using iter() ensures the ModelIterable is resumed with the next chunk.
        self._number_returned = 0
        self._in_iterator = True
        try:
            qs_iter = iter(self.queryset.iterator(chunk_size=self.sql_chunk_size))

            # Keep fetching chunks
            while True:
                instances = list(islice(qs_iter, self.chunk_size))
                if not instances:
                    break

                # Perform prefetches on this chunk:
                if self.queryset._prefetch_related_lookups:
                    self._add_prefetches(instances)

                # And return to parent loop
                yield from instances
                self._number_returned += len(instances)
        finally:
            self._in_iterator = False

    @property
    def number_returned(self) -> int:
        """Tell how many objects the iterator processed"""
        if self._in_iterator:
            raise RuntimeError("Can't read number of returned results during iteration")
        return self._number_returned

    def _add_prefetches(self, instances: list[M]):
        """Merge the prefetched objects for this batch with the model instances."""
        if self._fk_caches:
            # Make sure prefetch_related_objects() doesn't have
            # to fetch items again that infrequently changes.
            all_restored = self._restore_caches(instances)
            if all_restored:
                return

        # Reuse the Django machinery for retrieving missing sub objects.
        # and analyse the ForeignKey caches to allow faster prefetches next time
        models.prefetch_related_objects(
            instances, *self.queryset._prefetch_related_lookups
        )
        self._persist_prefetch_cache(instances)

    def _persist_prefetch_cache(self, instances):
        """Store the prefetched data so it can be applied to the next batch"""
        for instance in instances:
            for lookup, obj in instance._state.fields_cache.items():
                if obj is not None:
                    cache = self._fk_caches[lookup]
                    cache[obj.pk] = obj

    def _restore_caches(self, instances) -> bool:
        """Restore prefetched data to the new set of instances.
        This avoids unneeded prefetching of the same ForeignKey relation.
        """
        if not instances:
            return True
        if not self._fk_caches:
            return False

        all_restored = True

        for lookup, cache in self._fk_caches.items():
            field = instances[0]._meta.get_field(lookup)
            for instance in instances:
                id_value = getattr(instance, field.attname)
                if id_value is None:
                    continue

                obj = cache.get(id_value, None)
                if obj is not None:
                    instance._state.fields_cache[lookup] = obj
                else:
                    all_restored = False

        return all_restored



from __future__ import annotations

from functools import wraps
from itertools import chain
from typing import TYPE_CHECKING
from xml.etree.ElementTree import Element, QName

from gisserver.exceptions import ExternalParsingError


def expect_tag(namespace: str, *tag_names: str, leaf=False):
    """Validate whether a given tag is need"""
    valid_tags = {str(QName(namespace, name)) for name in tag_names}
    expect0 = str(QName(namespace, tag_names[0]))

    def _wrapper(func):
        @wraps(func)
        def _expect_tag_decorator(cls, element: Element, *args, **kwargs):
            if element.tag not in valid_tags:
                raise ExternalParsingError(
                    f"{cls.__name__} parser expects an <{expect0}> node, got <{element.tag}>"
                )
            if leaf and len(element):
                raise ExternalParsingError(
                    f"Unsupported child element for {element.tag} element: {element[0].tag}."
                )

            return func(cls, element, *args, **kwargs)

        return _expect_tag_decorator

    return _wrapper


def expect_children(min_child_nodes, *expect_types: str | type[BaseNode]):
    def _wrapper(func):
        @wraps(func)
        def _expect_children_decorator(cls, element: Element, *args, **kwargs):
            if len(element) < min_child_nodes:
                type_names = ", ".join(
                    sorted(
                        set(
                            chain.from_iterable(
                                (
                                    [child_type]
                                    if isinstance(child_type, str)
                                    else chain.from_iterable(
                                        sub_type.xml_tags
                                        for sub_type in child_type.__subclasses__()
                                    )
                                )
                                for child_type in expect_types
                            )
                        )
                    )
                )
                suffix = f" (possible tags: {type_names})" if type_names else ""
                raise ExternalParsingError(
                    f"<{element.tag}> should have {min_child_nodes} child nodes, "
                    f"got {len(element)}{suffix}"
                )

            return func(cls, element, *args, **kwargs)

        return _expect_children_decorator

    return _wrapper


def get_child(root, namespace, localname) -> Element:
    """Find the element using a fully qualified name."""
    return root.find(QName(namespace, localname).text)


def get_children(root, namespace, localname) -> list[Element]:
    """Find the element using a fully qualified name."""
    return root.findall(QName(namespace, localname).text)


def get_attribute(element: Element, name) -> str:
    """Resolve an attribute, raise an error when it's missing."""
    try:
        return element.attrib[name]
    except KeyError:
        raise ExternalParsingError(
            f"Element {element.tag} misses required attribute '{name}'"
        ) from None


def split_ns(tag_name: str) -> tuple[str | None, str]:
    """Split the element tag into the namespace and local name.
    The stdlib etree doesn't have the properties for this (lxml does).
    """
    if tag_name.startswith("{"):
        end = tag_name.index("}")
        return tag_name[1:end], tag_name[end + 1 :]
    else:
        return None, tag_name


if TYPE_CHECKING:
    from gisserver.parsers.base import BaseNode



"""Functions to be callable from fes."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Union

from django.contrib.gis.db.models import functions as gis
from django.db import models
from django.db.models import functions
from django.db.models.expressions import Combinable

from gisserver.exceptions import InvalidParameterValue
from gisserver.types import XsdTypes

__all__ = ["function_registry"]

FesArg = Union[Combinable, models.Q]
FesFunctionBody = Union[models.Func, Callable[..., models.Func]]


@dataclass(order=True)
class FesFunction:
    """Wrapper that defines a fes function, with type descriptions.
    This is also used to provide metadata to the GetCapabilities call.
    """

    #: Name of the function
    name: str

    #: The function body
    body: FesFunctionBody

    #: The argument names with their XSD types
    arguments: dict[str, XsdTypes] | None = None

    #: The XSD return type
    returns: XsdTypes | None = None

    def build_query(self, *expressions: FesArg) -> models.Func:
        """Build the query expression for the function."""
        if len(expressions) != len(self.arguments):
            # Avoid passing extra parameters to the function if those are not defined.
            raise TypeError(
                f'Invalid number of arguments for <fes:Function name="{self.name}">'
            )

        # Keyword arguments are avoided, since some Django SQL functions had
        # SQL-injection issues with those arguments. Only passing expressions for now.
        return self.body(*expressions)


class FesFunctionRegistry:
    """Registry of functions to be callable by <fes:Function>.

    The registered functions should be capable of running an SQL function.
    """

    def __init__(self):
        self.functions = {}

    def __bool__(self):
        return bool(self.functions)

    def __iter__(self):
        return iter(sorted(self.functions.values()))  # for template rendering

    def register(
        self,
        name=None,
        body: FesFunctionBody | None = None,
        *,
        arguments: dict[str, XsdTypes] | None = None,
        returns: XsdTypes | None = None,
    ):
        """Decorator to register a function.

        It's not recommended to register the Django Func objects directly,
        as the parameters are passed on from the client-side. Instead, create
        a wrapper function that enforces a controlled set of parameters.
        """

        def _wrapper(callable: FesFunctionBody):
            function = FesFunction(
                name=name or callable.__name__,
                body=callable,
                arguments=arguments,
                returns=returns,
            )
            self.functions[function.name] = function
            return callable

        if body is not None:
            # Usage as direct registration
            return _wrapper(body)
        else:
            # Usage as decorator
            return _wrapper

    def resolve_function(self, function_name) -> FesFunction:
        """Resole the function using it's name."""
        try:
            return self.functions[function_name]
        except KeyError:
            raise InvalidParameterValue(
                "filter", f"Unsupported function: {function_name}"
            ) from None


function_registry = FesFunctionRegistry()


# Register a set of default SQL functions.
# These are based on GeoServer:
# https://docs.geoserver.org/latest/en/user/filter/function_reference.html
# Not implemented:
# - Aggregates (like Collection_*)
# - Comparisons (which already has fes variants)

# -- strings

function_registry.register(
    "strConcat",
    functions.Concat,
    arguments={"string1": XsdTypes.string, "string2": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strToLowerCase",
    functions.Lower,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strToUpperCase",
    functions.Upper,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strTrim",
    functions.Trim,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strLength",
    functions.Length,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.integer,
)

function_registry.register(
    "length",
    functions.Length,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.integer,
)

# -- math numbers

function_registry.register(
    "abs",
    functions.Abs,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.double,
)

function_registry.register(
    "ceil",
    functions.Ceil,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.decimal,
)

function_registry.register(
    "floor",
    functions.Floor,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.decimal,
)

function_registry.register(
    "min",
    functions.Least,
    arguments={"value1": XsdTypes.double, "value2": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "max",
    functions.Greatest,
    arguments={"value1": XsdTypes.double, "value2": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "pow",
    functions.Power,
    arguments={"base": XsdTypes.double, "exponent": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "round",
    functions.Round,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.integer,
)

function_registry.register(
    "exp",
    functions.Exp,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "log",
    functions.Log,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "sqrt",
    functions.Sqrt,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

# -- math trigonometry

function_registry.register(
    "acos",
    functions.ACos,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "asin",
    functions.ASin,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "atan",
    functions.ATan,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "atan2",
    functions.ATan2,
    arguments={"x": XsdTypes.double, "y": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "cos",
    functions.Cos,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "sin",
    functions.Sin,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "tan",
    functions.Tan,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "pi",
    functions.Pi,
    returns=XsdTypes.double,
)

function_registry.register(
    "toDegrees",
    functions.Degrees,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "toRadians",
    functions.Radians,
    arguments={"degree": XsdTypes.double},
    returns=XsdTypes.double,
)

# -- geometric

function_registry.register(
    "Area",
    gis.Area,
    arguments={"geometry": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.double,
)

function_registry.register(
    "Centroid",
    gis.Centroid,
    arguments={"features": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.string,
)

function_registry.register(
    "Difference",
    gis.Difference,
    arguments={
        "geometry1": XsdTypes.gmlAbstractGeometryType,
        "geometry2": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.string,
)

function_registry.register(
    "distance",
    gis.Distance,
    arguments={
        "geometry1": XsdTypes.gmlAbstractGeometryType,
        "geometry2": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.double,
)

function_registry.register(
    "Envelope",
    gis.Envelope,
    arguments={"geometry": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.string,
)

function_registry.register(
    "Intersection",
    gis.Intersection,
    arguments={
        "geometry1": XsdTypes.gmlAbstractGeometryType,
        "geometry2": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.string,
)

function_registry.register(
    "Union",
    gis.Union,
    arguments={
        "geometry1": XsdTypes.gmlAbstractGeometryType,
        "geometry2": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.string,
)



from __future__ import annotations

from collections import deque
from typing import Iterable, cast

from gisserver.features import FeatureType
from gisserver.operations.base import WFSMethod
from gisserver.types import XsdComplexType

from .base import OutputRenderer
from .buffer import StringBuffer


class XMLSchemaRenderer(OutputRenderer):
    """Output rendering for DescribeFeatureType.

    This renders a valid XSD schema that describes the data type of the feature.
    """

    content_type = "application/gml+xml; version=3.2"  # mandatory for WFS

    def __init__(self, method: WFSMethod, feature_types: list[FeatureType]):
        """Overwritten method to handle the DescribeFeatureType context."""
        self.server_url = method.view.server_url
        self.app_xml_namespace = method.view.xml_namespace
        self.feature_types = feature_types

    def render_stream(self):
        # For now, all features have the same XML namespace despite allowing
        # the prefixes to be different.
        xmlns_features = "\n   ".join(
            f'xmlns:{p}="{self.app_xml_namespace}"'
            for p in sorted(
                {feature_type.xml_prefix for feature_type in self.feature_types}
            )
        )

        output = StringBuffer()
        output.write(
            f"""<?xml version='1.0' encoding="UTF-8" ?>
<schema
   xmlns="http://www.w3.org/2001/XMLSchema"
   {xmlns_features}
   xmlns:gml="http://www.opengis.net/gml/3.2"
   targetNamespace="{self.app_xml_namespace}"
   elementFormDefault="qualified" version="0.1">

"""
        )
        output.write(self.render_imports())
        output.write("\n")

        for feature_type in self.feature_types:
            output.write(self.render_feature_type(feature_type))

        output.write("</schema>\n")
        return output.getvalue()

    def render_imports(self):
        return (
            '  <import namespace="http://www.opengis.net/gml/3.2"'
            ' schemaLocation="http://schemas.opengis.net/gml/3.2.1/gml.xsd" />\n'
        )

    def render_feature_type(self, feature_type: FeatureType):
        output = StringBuffer()
        xsd_type: XsdComplexType = feature_type.xsd_type

        # This declares the that a top-level <app:featureName> is a class of a type.
        output.write(
            f'  <element name="{feature_type.name}"'
            f' type="{xsd_type}" substitutionGroup="gml:AbstractFeature" />\n\n'
        )

        # Next, the complexType is rendered that defines the element contents.
        # Next, the complexType(s) are rendered that defines the element contents.
        # In case any fields are expanded (hence become sub-types), these are also included.
        output.write(self.render_complex_type(xsd_type))
        for complex_type in self._get_complex_types(xsd_type):
            output.write(self.render_complex_type(complex_type))

        return output.getvalue()

    def render_complex_type(self, complex_type: XsdComplexType):
        """Write the definition of a single class."""
        output = StringBuffer()
        output.write(
            f'  <complexType name="{complex_type.name}">\n'
            "    <complexContent>\n"
            f'      <extension base="{complex_type.base}">\n'
            "        <sequence>\n"
        )

        for xsd_element in complex_type.elements:
            output.write(f"          {xsd_element}\n")

        output.write(
            "        </sequence>\n"
            "      </extension>\n"
            "    </complexContent>\n"
            "  </complexType>\n\n"
        )
        return output.getvalue()

    def _get_complex_types(self, root: XsdComplexType) -> Iterable[XsdComplexType]:
        """Find all fields that reference to complex types, including nested elements."""
        elements = deque(root.elements)
        complex_types = {}

        # Walk through next types, unless they are already seen.
        # No recursion is used here to handle circular loops of types.
        while elements:
            element = elements.popleft()
            element_type = element.type
            if element_type.is_complex_type and element_type.name not in complex_types:
                # ComplexType was not seen before, register it.
                # It's members are added to the deque for recursive processing.
                xsd_element_type = cast(XsdComplexType, element_type)
                complex_types[element_type.name] = xsd_element_type
                elements.extend(xsd_element_type.elements)

        # Present in a consistent order
        return complex_types.values()



from __future__ import annotations

import operator
from functools import reduce

from django.conf import settings
from django.contrib.gis.db.models.fields import BaseSpatialField
from django.contrib.gis.db.models.lookups import DWithinLookup
from django.db import models
from django.db.models import Q, QuerySet, lookups
from django.db.models.expressions import Combinable

from gisserver.features import FeatureType

from . import expressions, sorting


class CompiledQuery:
    """Intermediate data for translating FES queries to Django.

    This class contains all data from the ``<fes:Filter>`` object in a model
    that can be translated to a django QuerySet.

    This object is passed though all build_...() methods,
    so it can be used to add extra lookups and annotations.
    """

    def __init__(
        self,
        feature_type: FeatureType | None = None,
        using: str | None = None,
        lookups: list[Q] | None = None,
        typed_lookups: dict[str, list[Q]] | None = None,
        annotations: dict[str, Combinable | Q] | None = None,
    ):
        """The init method is typically used only in unit tests."""
        self.feature_type = feature_type
        self.using = using
        self.lookups = lookups or []
        self.typed_lookups = typed_lookups or {}
        self.annotations = annotations or {}
        self.aliases = 0
        self.extra_lookups: list[Q] = []
        self.ordering: list[str] = []
        self.is_empty = False
        self.distinct = False

    def add_annotation(self, value: Combinable | Q) -> str:
        """Create a named-alias for a function/Q object.
        This alias can be used in a comparison, where expressions are used as left-hand-side.
        """
        self.aliases += 1
        name = f"a{self.aliases}"
        self.annotations[name] = value
        return name

    def add_distinct(self):
        self.distinct = True

    def add_lookups(self, q_object: Q, type_name: str | None = None):
        """Register an extra 'WHERE' clause of the query.
        This is used for comparisons, ID selectors and other query types.
        """
        if not isinstance(q_object, Q):
            raise TypeError()

        if type_name is not None:
            if type_name not in self.typed_lookups:
                self.typed_lookups[type_name] = []
            self.typed_lookups[type_name].append(q_object)
        else:
            self.lookups.append(q_object)

    def add_extra_lookup(self, q_object: Q):
        """Temporary stash an extra lookup that the expression can't return yet.
        This is used for XPath selectors that also filter on attributes,
        e.g. "element[@attr=..]/child". The attribute lookup is processed as another filter.
        """
        if not isinstance(q_object, Q):
            raise TypeError()
        self.extra_lookups.append(q_object)

    def add_sort_by(self, sort_by: sorting.SortBy):
        """Read the desired result ordering from a ``<fes:SortBy>`` element."""
        self.ordering += sort_by.build_ordering(self.feature_type)

    def add_value_reference(self, value_reference: expressions.ValueReference) -> str:
        """Add a reference that should be returned by the query.

        This includes the XPath expression to the query, in case that adds
        extra lookups. The name (or alias) is returned that can be used in the
         ``queryset.values()`` result. This is needed to support cases like
        these in the future: ``addresses/Address[street="Oxfordstrasse"]/number``
        """
        return value_reference.build_rhs(self)

    def apply_extra_lookups(self, comparison: Q) -> Q:
        """Combine stashed lookups with the provided Q object.

        This is called for functions that compile a "Q" object.
        In case a node added extra lookups (for attributes), these are combined here
        with the actual comparison.
        """
        if not self.extra_lookups:
            return comparison

        # The extra lookups are used for XPath queries such as "/node[@attr=..]/foo".
        # A <ValueReference> with such lookup also requires to limit the filtered results,
        # in addition to the comparison operator code that is wrapped up here.
        result = reduce(operator.and_, [comparison] + self.extra_lookups)
        self.extra_lookups.clear()
        return result

    def mark_empty(self):
        """Mark as returning no results."""
        self.is_empty = True

    def filter_queryset(
        self, queryset: QuerySet, feature_type: FeatureType
    ) -> QuerySet:
        """Apply the filters and lookups to the queryset.

        :param queryset: The queryset to filter.
        :param feature_type: The feature type that the queryset originated from.
        """
        if self.is_empty:
            return queryset.none()

        if self.extra_lookups:
            # Each time an expression node calls add_extra_lookup(),
            # the parent should have used apply_extra_lookups()
            raise RuntimeError("apply_extra_lookups() was not called")

        # All are applied at once.
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        lookups = self.lookups
        try:
            lookups += self.typed_lookups[feature_type.name]
        except KeyError:
            pass

        if lookups:
            queryset = queryset.filter(*lookups)

        if self.ordering:
            queryset = queryset.order_by(*self.ordering)

        if self.distinct:
            queryset = queryset.distinct()

        return queryset

    def __repr__(self):
        return (
            "<CompiledQuery"
            f" annotations={self.annotations!r},"
            f" lookups={self.lookups!r},"
            f" typed_lookups={self.typed_lookups!r}>"
        )

    def __eq__(self, other):
        """For pytest comparisons."""
        if isinstance(other, CompiledQuery):
            return (
                other.lookups == self.lookups
                and other.typed_lookups == self.typed_lookups
                and other.annotations == self.annotations
            )
        else:
            return NotImplemented


@models.CharField.register_lookup
@models.TextField.register_lookup
@models.ForeignObject.register_lookup
class FesLike(lookups.Lookup):
    """Allow fieldname__fes_like=... lookups in querysets."""

    lookup_name = "fes_like"

    def as_sql(self, compiler, connection):
        """Generate the required SQL."""
        # lhs = "table"."field"
        # rhs = %s
        # lhs_params = []
        # lhs_params = ["prep-value"]
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        return f"{lhs} LIKE {rhs}", lhs_params + rhs_params

    def get_db_prep_lookup(self, value, connection):
        """This expects that the right-hand-side already has wildcard characters."""
        return "%s", [value]


@models.Field.register_lookup
@models.ForeignObject.register_lookup
class FesNotEqual(lookups.Lookup):
    """Allow fieldname__fes_notequal=... lookups in querysets."""

    lookup_name = "fes_notequal"

    def as_sql(self, compiler, connection):
        """Generate the required SQL."""
        lhs, lhs_params = self.process_lhs(compiler, connection)  # = (table.field, %s)
        rhs, rhs_params = self.process_rhs(compiler, connection)  # = ("prep-value", [])
        return f"{lhs} != {rhs}", (lhs_params + rhs_params)


@BaseSpatialField.register_lookup
class FesBeyondLookup(DWithinLookup):
    """Based on the FES 2.0.3 corrigendum:

    DWithin(A,B,d) = Distance(A,B) < d
    Beyond(A,B,d) = Distance(A,B) > d

    See: https://docs.opengeospatial.org/is/09-026r2/09-026r2.html#61
    """

    lookup_name = "fes_beyond"
    sql_template = "NOT %(func)s(%(lhs)s, %(rhs)s, %(value)s)"

    def get_rhs_op(self, connection, rhs):
        # Allow the SQL $(func)s to be different from the ORM lookup name.
        # This uses ST_DWithin() on PostGIS
        return connection.ops.gis_operators["dwithin"]


if "django.contrib.postgres" in settings.INSTALLED_APPS:
    from django.contrib.postgres.fields import ArrayField

    class ArrayAnyMixin:
        any_operators = {
            "exact": "= ANY(%s)",
            "ne": "!= ANY(%s)",
            "gt": "< ANY(%s)",
            "gte": "<= ANY(%s)",
            "lt": "> ANY(%s)",
            "lte": ">= ANY(%s)",
        }

        def as_sql(self, compiler, connection):
            # For the ANY() comparison, the filter operands need to be reversed.
            # So instead of "field < value", it becomes "value > ANY(field)
            lhs_sql, lhs_params = self.process_lhs(compiler, connection)
            rhs_sql, rhs_params = self.process_rhs(compiler, connection)
            lhs_sql = self.get_rhs_op(connection, lhs_sql)
            return f"{rhs_sql} {lhs_sql}", (rhs_params + lhs_params)

        def get_rhs_op(self, connection, rhs):
            return self.any_operators[self.lookup_name] % rhs

    def _register_any_lookup(base: type[lookups.BuiltinLookup]):
        """Register array lookups under a different name."""
        cls = type(f"FesArrayAny{base.__name__}", (ArrayAnyMixin, base), {})
        ArrayField.register_lookup(cls, lookup_name=f"fes_any{base.lookup_name}")

    _register_any_lookup(lookups.Exact)
    _register_any_lookup(lookups.Exact)
    _register_any_lookup(lookups.GreaterThan)
    _register_any_lookup(lookups.GreaterThanOrEqual)
    _register_any_lookup(lookups.LessThan)
    _register_any_lookup(lookups.LessThanOrEqual)

    @ArrayField.register_lookup
    class FesArrayAnyNotEqual(FesNotEqual):
        """Inequality test for a single item in the array"""

        lookup_name = "fes_anynotequal"

        def as_sql(self, compiler, connection):
            """Generate the required SQL."""
            lhs, lhs_params = self.process_lhs(compiler, connection)
            rhs, rhs_params = self.process_rhs(compiler, connection)
            return f"{rhs} != ANY({lhs})", (rhs_params + lhs_params)



from __future__ import annotations

from enum import Enum
from xml.etree.ElementTree import Element, QName

from gisserver.exceptions import ExternalParsingError

from .tags import split_ns


class TagNameEnum(Enum):
    """An enumeration of tag names.

    All enumerations that represent tag names inherit from this.
    Each member name should be exactly the XML tag that it refers to.
    """

    @classmethod
    def from_xml(cls, element: Element):
        """Cast the element tag name into the enum member"""
        ns, localname = split_ns(element.tag)
        return cls[localname]

    @classmethod
    def _missing_(cls, value):
        raise NotImplementedError(
            f"<{value}> is not registered as valid {cls.__name__}"
        )

    def __repr__(self):
        # Make repr(filter) easier to copy-paste
        return f"{self.__class__.__name__}.{self.name}"


class BaseNode:
    """The base node for all classes that represent an XML tag."""

    xml_ns = None
    xml_tags = []

    def __init_subclass__(cls):
        cls.xml_tags = []

    @classmethod
    def from_xml(cls, element: Element):
        raise NotImplementedError(
            f"{cls.__name__}.from_xml() is not implemented to parse <{element.tag}>"
        )

    @classmethod
    def from_child_xml(cls, element: Element) -> BaseNode:
        """By default, the node attempts to locate a child-class from the registry.
        The individual tags should override `from_xml()` to return their own type.
        """
        return tag_registry.from_child_xml(element, allowed_types=(cls,))


class TagRegistry:
    """Registration of all classes that can parse XML nodes.

    The same class can be registered multiple times for different tag names.
    """

    parsers: dict[str, type[BaseNode]]

    def __init__(self):
        self.parsers = {}

    def register(self, name=None, namespace=None, hidden=False):
        """Decorator to register an class as XML node parser.

        This registers the decorated class as the designated parser
        for a specific XML tag.

        Usage:

            @dataclass
            @tag_registry.register()
            class SomeXmlTag(BaseNode):
                xml_ns = FES

                @classmethod
                def from_xml(cls, element: Element):
                    return cls(
                        ...
                    )

        """

        def _dec(sub_class: type[BaseNode]) -> type[BaseNode]:
            if sub_class.xml_ns is None:
                raise RuntimeError(f"{sub_class.__name__}.xml_ns should be set")

            localname = name or sub_class.__name__
            qname = QName(namespace or sub_class.xml_ns, localname)

            self.parsers[qname] = sub_class  # Track this parser to resolve the tag.
            if not hidden:
                sub_class.xml_tags.append(name)  # Allow fetching all names later
            return sub_class  # allow decorator usage

        return _dec

    def register_names(self, names: type[TagNameEnum], namespace=None):
        """Decorator to register the 'tag_class' as the
        parser-backend for all member-names in this enum.
        """

        def _dec(sub_class: type[BaseNode]):
            # Looping over _member_names_ will skip aliased items (like BBOX/Within)
            for member_name in names.__members__.keys():
                self.register(name=member_name, namespace=namespace)(sub_class)
            return sub_class

        return _dec

    def from_child_xml(self, element: Element, allowed_types=None) -> BaseNode:
        """Convert the element into a Python class.

        This locates the parser from the registered tag.
        It's assumed that the tag has an "from_xml()" method too.
        """
        try:
            real_cls = self.resolve_class(element.tag)
        except ExternalParsingError as e:
            if allowed_types:
                # Show better exception message
                types = ", ".join(c.__name__ for c in allowed_types)
                raise ExternalParsingError(f"{e}, expected one of: {types}") from None

            raise

        # Check whether the resolved class is indeed a valid option here.
        if allowed_types is not None and not issubclass(real_cls, allowed_types):
            types = ", ".join(c.__name__ for c in allowed_types)
            raise ExternalParsingError(
                f"Unexpected {real_cls.__name__} for <{element.tag}> node, "
                f"expected one of: {types}"
            )

        return real_cls.from_xml(element)

    def resolve_class(self, tag_name) -> type[BaseNode]:
        # Resolve the dataclass using the tag name
        try:
            return self.parsers[tag_name]
        except KeyError:
            raise ExternalParsingError(f"Unsupported tag: <{tag_name}>") from None


tag_registry = TagRegistry()



"""These classes map to the FES 2.0 specification for operators.
The class names and attributes are identical to those in the FES spec.
"""
from __future__ import annotations

import operator
from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum
from functools import reduce
from itertools import groupby
from typing import Any, Union
from xml.etree.ElementTree import Element, QName

from django.conf import settings
from django.contrib.gis import measure
from django.db.models import Q
from django.utils.functional import cached_property

from gisserver.exceptions import ExternalParsingError, OperationProcessingFailed
from gisserver.parsers import gml
from gisserver.parsers.base import BaseNode, TagNameEnum, tag_registry
from gisserver.parsers.tags import expect_children, expect_tag, get_attribute, get_child
from gisserver.types import FES20

from .expressions import Expression, Literal, RhsTypes, ValueReference
from .identifiers import Id
from .query import CompiledQuery

SpatialDescription = Union[gml.GM_Object, gml.GM_Envelope, ValueReference]
TemporalOperand = Union[gml.TM_Object, ValueReference]


# Define interface for any class that has "build_rhs()"
try:
    from typing import Protocol
except ImportError:
    HasBuildRhs = Any  # Python 3.7 and below
else:

    class HasBuildRhs(Protocol):
        def build_rhs(self, compiler) -> RhsTypes:
            ...


if "django.contrib.postgres" in settings.INSTALLED_APPS:
    # Comparisons with array fields go through a separate ORM lookup expression,
    # so these can check whether ANY element matches in the array.
    # This gives consistency between other repeated elements (e.g. M2M, reverse FK)
    # where the whole object is returned when one of the sub-objects match.
    ARRAY_LOOKUPS = {
        "exact": "fes_anyexact",
        "fes_notequal": "fes_anynotequal",
        "lt": "fes_anylt",
        "lte": "fes_anylte",
        "gt": "fes_anygt",
        "gte": "fes_anygte",
    }
else:
    ARRAY_LOOKUPS = None


class MatchAction(Enum):
    """Values for the 'matchAction' attribute of the BinaryComparisonOperator."""

    All = "All"
    Any = "Any"
    One = "One"

    def __repr__(self):
        # Make repr(filter) easier to copy-paste
        return f"{self.__class__.__name__}.{self.name}"


class BinaryComparisonName(TagNameEnum):
    """XML tag names for value comparisons.
    Values are the used field lookup.
    """

    PropertyIsEqualTo = "exact"
    PropertyIsNotEqualTo = "fes_notequal"  # custom FesNotEqualTo lookup
    PropertyIsLessThan = "lt"
    PropertyIsGreaterThan = "gt"
    PropertyIsLessThanOrEqualTo = "lte"
    PropertyIsGreaterThanOrEqualTo = "gte"


class DistanceOperatorName(TagNameEnum):
    """XML tag names for distance operators"""

    Beyond = "fes_beyond"  # using __distance_gt=.. would be slower.
    DWithin = "dwithin"  # ST_DWithin uses indices, distance_lte does not.


class SpatialOperatorName(TagNameEnum):
    """XML tag names for geometry operators"""

    # (A Within B) implies that (B Contains A)

    # BBOX can either be implemented using bboverlaps (more efficient), or the
    # more correct "intersects" option (e.g. a line near the box would match otherwise).
    BBOX = "intersects"  # ISO version: "NOT DISJOINT"
    Equals = "equals"  # Test whether t geometries are topologically equal
    Disjoint = "disjoint"  # Tests whether two geometries are disjoint (do not interact)
    Intersects = "intersects"  # Tests whether two geometries intersect
    Touches = "touches"  # Tests whether two geometries touch
    Crosses = "crosses"  # Tests whether two geometries cross
    Within = "within"  # Tests whether a geometry is within another one
    Contains = "contains"  # Tests whether a geometry contains another one
    Overlaps = "overlaps"  # Test whether two geometries overlap


class TemporalOperatorName(TagNameEnum):
    """XML tag names for datetime operators.

    Explanation here: http://old.geotools.org/Temporal-Filters_211091519.html
    and: https://github.com/geotools/geotools/wiki/temporal-filters
    """

    After = "after"
    Before = "before"
    Begins = "begins"
    BegunBy = "begunby"
    TContains = "tcontains"
    TEquals = "tequals"
    TOverlaps = "toverlaps"
    During = "during"
    Meets = "meets"
    OverlappedBy = "overlappedby"
    MetBy = "metby"
    EndedBy = "endedby"
    AnyInteracts = "anyinteracts"


class BinaryLogicType(TagNameEnum):
    """XML tag names for the BinaryLogicOperator."""

    And = operator.and_
    Or = operator.or_


class UnaryLogicType(TagNameEnum):
    """XML tag names for the UnaryLogicOperator."""

    Not = operator.inv


@dataclass
class Measure(BaseNode):
    """The <fes:Distance uom="...> element."""

    xml_ns = FES20

    value: Decimal
    uom: str  # Unit of measurement, fes20:UomSymbol | fes20:UomURI

    @classmethod
    @expect_tag(FES20, "Distance")
    def from_xml(cls, element: Element):
        return cls(value=Decimal(element.text), uom=get_attribute(element, "uom"))

    def build_rhs(self, compiler) -> measure.Distance:
        return measure.Distance(default_unit=self.uom, **{self.uom: self.value})


class Operator(BaseNode):
    """Abstract base class, as defined by FES spec."""

    xml_ns = FES20

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        raise NotImplementedError(
            f"Using {self.__class__.__name__} is not supported yet."
        )


@dataclass
class IdOperator(Operator):
    """List of ResourceId objects"""

    id: list[Id]

    @property
    def type_names(self) -> list[str]:
        """Provide a list of all type names accessed by this operator"""
        return [
            type_name for type_name in self.grouped_ids.keys() if type_name is not None
        ]

    @cached_property
    def grouped_ids(self) -> dict[str, list[Id]]:
        # For itertools.groupby(), items have to be sorted first.
        ids = sorted(self.id, key=operator.attrgetter("rid"))
        return {
            type_name: list(items)
            for type_name, items in groupby(ids, key=operator.attrgetter("type_name"))
        }

    def build_query(self, compiler):
        """Generate the ID lookup query.

        As these identifiers also reference the type name, no Q-object is
        returned. The lookups are directly added to the fes query object.
        """
        # When ResourceId has no type_name associated, this means that the
        # parsing was invalid. Instead of raising an error, empty results are returned.
        if len(self.grouped_ids) == 1 and self.grouped_ids.get(None):
            compiler.mark_empty()
            return

        for type_name, items in self.grouped_ids.items():
            ids_subset = reduce(
                operator.or_, [id.build_query(compiler=None) for id in items]
            )
            compiler.add_lookups(ids_subset, type_name=type_name)


class NonIdOperator(Operator):
    """Abstract base class, as defined by FES spec.

    This is used for nearly all operators,
    except those that have <fes:ResourceId> elements as children.
    """

    _source = None
    allow_geometries = False

    def build_compare(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: Expression | RhsTypes,
    ) -> Q:
        """Use the value in comparison with some other expression.

        This calls build_lhs() and build_rhs() on the expressions.
        """
        # lhs and rhs are allowed to be reversed. However, the SQL compiler
        # works much simpler when Django can predict the actual data type.
        if isinstance(lhs, Literal) and isinstance(rhs, ValueReference):
            lhs, rhs = rhs, lhs

        if compiler.feature_type is not None:
            lookup = self.validate_comparison(compiler, lhs, lookup, rhs)

        # Build Django Q-object
        lhs = lhs.build_lhs(compiler)

        if isinstance(rhs, (Expression, gml.GM_Object)):
            rhs = rhs.build_rhs(compiler)

        comparison = Q(**{f"{lhs}__{lookup}": rhs})
        return compiler.apply_extra_lookups(comparison)

    def validate_comparison(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: Expression | RhsTypes,
    ):
        """Validate whether a given comparison is even possible.
        Where needed, the lhs/rhs are already ordered in a logical sequence.
        """
        if isinstance(lhs, ValueReference):
            xsd_element = compiler.feature_type.resolve_element(lhs.xpath).child
            tag = self._source if self._source is not None else None

            # e.g. deny <PropertyIsLessThanOrEqualTo> against <gml:boundedBy>
            if xsd_element.is_geometry and not self.allow_geometries:
                raise OperationProcessingFailed(
                    "filter",
                    f"Operator '{tag}' does not support comparing"
                    f" geometry properties: '{xsd_element.xml_name}'.",
                    status_code=400,  # not HTTP 500 here. Spec allows both.
                )

            if isinstance(rhs, Literal):
                # Since the element is resolved, inform the Literal how to parse the value.
                # This avoids various validation errors along the path.
                rhs.bind_type(xsd_element.type)

                # When a common case of value comparison is done, the inputs
                # can be validated before the ORM query is constructed.
                xsd_element.validate_comparison(rhs.raw_value, lookup=lookup, tag=tag)

            # Checking scalar values against array fields will fail.
            # However, to make the queries consistent with other unbounded types (i.e. M2M fields),
            # it makes sense to return an object when *one* entry in the array matches.
            if xsd_element.is_array:
                try:
                    return ARRAY_LOOKUPS[lookup]
                except KeyError:
                    raise OperationProcessingFailed(
                        "filter",
                        f"Operator '{tag}' is not supported for "
                        f"the '{xsd_element.name}' property.",
                        status_code=400,  # not HTTP 500 here. Spec allows both.
                    ) from None

        return lookup

    def build_compare_between(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: tuple[HasBuildRhs, HasBuildRhs],
    ) -> Q:
        """Use the value in comparison with 2 other values (e.g. between query)"""
        if compiler.feature_type is not None:
            self.validate_comparison(compiler, lhs, lookup, rhs[0])
            self.validate_comparison(compiler, lhs, lookup, rhs[1])

        field_name = lhs.build_lhs(compiler)
        comparison = Q(
            **{
                f"{field_name}__{lookup}": (
                    rhs[0].build_rhs(compiler),
                    rhs[1].build_rhs(compiler),
                )
            }
        )
        return compiler.apply_extra_lookups(comparison)


class SpatialOperator(NonIdOperator):
    """Abstract base class, as defined by FES spec."""


@dataclass
@tag_registry.register_names(DistanceOperatorName)  # <Beyond>, <DWithin>
class DistanceOperator(SpatialOperator):
    """Comparing the distance to a geometry."""

    allow_geometries = True  # override static attribute

    valueReference: ValueReference
    operatorType: DistanceOperatorName
    geometry: gml.GM_Object
    distance: Measure
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(3, ValueReference, gml.GM_Object, Measure)
    def from_xml(cls, element: Element):
        geometries = gml.find_gml_nodes(element)
        if not geometries:
            raise ExternalParsingError(f"Missing gml element in <{element.tag}>")
        elif len(geometries) > 1:
            raise ExternalParsingError(
                f"Multiple gml elements found in <{element.tag}>"
            )

        return cls(
            valueReference=ValueReference.from_xml(
                get_child(element, FES20, "ValueReference")
            ),
            operatorType=DistanceOperatorName.from_xml(element),
            geometry=gml.parse_gml_node(geometries[0]),
            distance=Measure.from_xml(get_child(element, FES20, "Distance")),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        return self.build_compare_between(
            compiler,
            lhs=self.valueReference,
            lookup=self.operatorType.value,
            rhs=(self.geometry, self.distance),
        )


@dataclass
@tag_registry.register_names(SpatialOperatorName)  # <BBOX>, <Equals>, ...
class BinarySpatialOperator(SpatialOperator):
    """A comparison of geometries using 2 values, e.g. A Within B."""

    allow_geometries = True  # override static attribute

    operatorType: SpatialOperatorName
    operand1: ValueReference | None
    operand2: SpatialDescription
    _source: str | None = field(compare=False, default=None)

    @classmethod
    def from_xml(cls, element: Element):
        operator_type = SpatialOperatorName.from_xml(element)
        if operator_type is SpatialOperatorName.BBOX and len(element) == 1:
            # For BBOX, the geometry operator is optional
            ref = None
            geo = element[0]
        else:
            if len(element) != 2:
                raise ExternalParsingError(f"{element.tag} should have 2 operators")
            ref, geo = list(element)

        return cls(
            operatorType=operator_type,
            operand1=ValueReference.from_xml(ref) if ref is not None else None,
            operand2=tag_registry.from_child_xml(
                geo, allowed_types=SpatialDescription.__args__  # get_args() in 3.8
            ),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        operant1 = self.operand1
        if operant1 is None:
            operant1 = ValueReference(xpath=compiler.feature_type.geometry_field.name)

        return self.build_compare(
            compiler,
            lhs=operant1,
            lookup=self.operatorType.value,
            rhs=self.operand2,
        )


@dataclass
@tag_registry.register_names(TemporalOperatorName)  # <After>, <Before>, ...
class TemporalOperator(NonIdOperator):
    """Comparisons with dates"""

    operatorType: TemporalOperatorName
    operand1: ValueReference
    operand2: TemporalOperand
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(2, ValueReference, TemporalOperand)
    def from_xml(cls, element: Element):
        return cls(
            operatorType=TemporalOperatorName.from_xml(element),
            operand1=ValueReference.from_xml(element[0]),
            operand2=tag_registry.from_child_xml(
                element[1], allowed_types=TemporalOperand.__args__  # get_args() in 3.8
            ),
            _source=element.tag,
        )


class ComparisonOperator(NonIdOperator):
    """Base class for comparisons"""

    # Start counting fresh here, to collect the capabilities
    # that are listed in the <fes20:ComparisonOperators> node:
    xml_tags = []


@dataclass
@tag_registry.register_names(BinaryComparisonName)  # <PropertyIs...>
class BinaryComparisonOperator(ComparisonOperator):
    """A comparison between 2 values, e.g. A == B."""

    operatorType: BinaryComparisonName
    expression: tuple[Expression, Expression]
    matchCase: bool = True
    matchAction: MatchAction = MatchAction.Any
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(2, Expression, Expression)
    def from_xml(cls, element: Element):
        return cls(
            operatorType=BinaryComparisonName.from_xml(element),
            expression=(
                Expression.from_child_xml(element[0]),
                Expression.from_child_xml(element[1]),
            ),
            matchCase=element.get("matchCase", True),
            matchAction=MatchAction(
                element.get("matchAction", default=MatchAction.Any)
            ),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        lhs, rhs = self.expression
        return self.build_compare(
            compiler, lhs=lhs, lookup=self.operatorType.value, rhs=rhs
        )


@dataclass
@tag_registry.register("PropertyIsBetween")
class BetweenComparisonOperator(ComparisonOperator):
    """Check whether a value is between two elements."""

    expression: Expression
    lowerBoundary: Expression
    upperBoundary: Expression
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(3, Expression, "LowerBoundary", "UpperBoundary")
    def from_xml(cls, element: Element):
        if element[1].tag != QName(FES20, "LowerBoundary") or element[2].tag != QName(
            FES20, "UpperBoundary"
        ):
            raise ExternalParsingError(
                f"{element.tag} should have 3 child nodes: "
                f"(expression), <LowerBoundary>, <UpperBoundary>"
            )

        lower = get_child(element, FES20, "LowerBoundary")
        upper = get_child(element, FES20, "UpperBoundary")

        if len(lower) != 1:
            raise ExternalParsingError(
                f"{lower.tag} should have 1 expression child node"
            )
        if len(upper) != 1:
            raise ExternalParsingError(
                f"{upper.tag} should have 1 expression child node"
            )

        return cls(
            expression=Expression.from_child_xml(element[0]),
            lowerBoundary=Expression.from_child_xml(lower[0]),
            upperBoundary=Expression.from_child_xml(upper[0]),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        return self.build_compare_between(
            compiler,
            lhs=self.expression,
            lookup="range",
            rhs=(self.lowerBoundary, self.upperBoundary),
        )


@dataclass
@tag_registry.register("PropertyIsLike")
class LikeOperator(ComparisonOperator):
    """Perform wildcard matching."""

    expression: tuple[Expression, Expression]
    wildCard: str
    singleChar: str
    escapeChar: str
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(2, Expression, Expression)
    def from_xml(cls, element: Element):
        return cls(
            expression=(
                Expression.from_child_xml(element[0]),
                Expression.from_child_xml(element[1]),
            ),
            # These attributes are required by the WFS spec:
            wildCard=get_attribute(element, "wildCard"),
            singleChar=get_attribute(element, "singleChar"),
            escapeChar=get_attribute(element, "escapeChar"),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        lhs, rhs = self.expression
        if isinstance(rhs, Literal):
            value = str(rhs.value)  # value could be auto-casted to int.

            # Not using r"\" here as that is a syntax error.
            if self.escapeChar != "\\":
                value = value.replace("\\", "\\\\").replace(self.escapeChar, "\\")
            if self.wildCard != "%":
                value = value.replace("%", r"\%").replace(self.wildCard, "%")
            if self.singleChar != "_":
                value = value.replace("_", r"\_").replace(self.singleChar, "_")

            rhs = Literal(raw_value=value)
        else:
            raise ExternalParsingError(
                f"Expected a literal value for the {self.tag} operator."
            )

        # Use the FesLike lookup
        return self.build_compare(compiler, lhs=lhs, lookup="fes_like", rhs=rhs)


@dataclass
@tag_registry.register("PropertyIsNil")
class NilOperator(ComparisonOperator):
    """Check whether the value evaluates to null/None.
    If the WFS returned a property element with <tns:p xsi:nil='true'>, this returns true.
    """

    expression: Expression | None
    nilReason: str
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(1, Expression)
    def from_xml(cls, element: Element):
        return cls(
            expression=Expression.from_child_xml(element[0]) if element else None,
            nilReason=element.get("nilReason"),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        # Any value that evaluates to None is returned as 'xs:nil' in our output.
        return self.build_compare(
            compiler, lhs=self.expression, lookup="isnull", rhs=True
        )


@dataclass
@tag_registry.register("PropertyIsNull")
class NullOperator(ComparisonOperator):
    """Check whether the property exists.
    If the WFS would not return the property element <tns:p>, this returns true.
    """

    expression: Expression
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(1, Expression)
    def from_xml(cls, element: Element):
        return cls(
            expression=Expression.from_child_xml(element[0]), _source=element.tag
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        # For now, the implementation is identical to PropertyIsNil.
        # According to the WFS spec, this should only be true when the element
        # is not returned at all (minOccurs=0).
        # TODO: this happens for maxOccurs=unbounded with a null value.
        return self.build_compare(
            compiler, lhs=self.expression, lookup="isnull", rhs=True
        )


class LogicalOperator(NonIdOperator):
    """Base class for AND, OR, NOT comparisons"""


@dataclass
@tag_registry.register("And")
@tag_registry.register("Or")
class BinaryLogicOperator(LogicalOperator):
    """Apply an AND or OR operator"""

    operands: list[NonIdOperator]
    operatorType: BinaryLogicType
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(2, NonIdOperator, NonIdOperator)
    def from_xml(cls, element: Element):
        return cls(
            operands=[NonIdOperator.from_child_xml(child) for child in element],
            operatorType=BinaryLogicType.from_xml(element),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        """Apply the AND/OR operation to the Q-object"""
        values = [q.build_query(compiler) for q in self.operands]
        return reduce(self.operatorType.value, values)


@dataclass
@tag_registry.register("Not")
class UnaryLogicOperator(LogicalOperator):
    """Apply a NOT operator"""

    operands: NonIdOperator
    operatorType: UnaryLogicType
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(1, NonIdOperator)
    def from_xml(cls, element: Element):
        return cls(
            operands=NonIdOperator.from_child_xml(element[0]),
            operatorType=UnaryLogicType.from_xml(element),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        """Apply the NOT operation to the Q-object"""
        return self.operatorType.value(self.operands.build_query(compiler))


class ExtensionOperator(NonIdOperator):
    """Base class for extensions to FES20"""



"""Output rendering logic for GeoJSON."""
from datetime import datetime
from decimal import Decimal
from typing import cast

import orjson
from django.conf import settings
from django.contrib.gis.db.models.functions import AsGeoJSON
from django.db import models
from django.utils.functional import Promise
from django.utils.timezone import utc

from gisserver import conf
from gisserver.db import get_db_geometry_target
from gisserver.features import FeatureType
from gisserver.geometries import CRS
from gisserver.types import XsdComplexType

from .base import OutputRenderer
from .buffer import BytesBuffer


def _json_default(obj):
    """Serialize non-built in values to JSON"""
    if isinstance(obj, (Decimal, Promise)):
        return str(obj)
    raise TypeError(f"Unable to serialize {obj.__class__.__name__} to JSON")


class GeoJsonRenderer(OutputRenderer):
    """Fast GeoJSON renderer, using a stream response.

    The complex encoding bits are handled by the C-library "orjson"
    and the geojson property of GEOSGeometry.

    NOTE: While Django has a GeoJSON serializer
    (see https://docs.djangoproject.com/en/3.0/ref/contrib/gis/serializers/),
    it does not offer streaming response handling.
    """

    content_type = "application/geo+json; charset=utf-8"
    content_disposition = 'inline; filename="{typenames} {page} {date}.geojson"'
    max_page_size = conf.GISSERVER_GEOJSON_MAX_PAGE_SIZE

    @classmethod
    def decorate_queryset(
        cls,
        feature_type: FeatureType,
        queryset: models.QuerySet,
        output_crs: CRS,
        **params,
    ):
        queryset = super().decorate_queryset(
            feature_type, queryset, output_crs, **params
        )

        # Other geometries can be excluded as these are not rendered by 'properties'
        other_geometries = [
            model_field.name
            for model_field in feature_type.geometry_fields
            if model_field is not feature_type.geometry_field
        ]
        if other_geometries:
            queryset = queryset.defer(*other_geometries)

        return queryset

    def render_stream(self):
        output = BytesBuffer()

        # Generate the header from a Python dict,
        # but replace the last "}" into a comma, to allow writing more
        header = self.get_header()

        # Have a temporary buffer that is written
        output.write(orjson.dumps(header)[:-1])
        output.write(b',\n  "features": [\n')

        # Flatten the results, they are not grouped in a second FeatureCollection
        is_first_collection = True
        for sub_collection in self.collection.results:
            if is_first_collection:
                is_first_collection = False
            else:
                output.write(b",\n")

            is_first = True
            for instance in sub_collection:
                if is_first:
                    is_first = False
                else:
                    output.write(b",\n")

                # The "properties" object is generated by orjson.dumps(),
                # while the "geometry" object uses the built-in 'GEOSGeometry.json' result.
                output.write(self.render_feature(sub_collection.feature_type, instance))

                # Only perform a 'yield' every once in a while,
                # as it goes back-and-forth for writing it to the client.
                if output.is_full():
                    yield output.flush()

        # Instead of performing an expensive .count() on the start of the page,
        # write this as a last field at the end of the response.
        # This still honors the WFS 30 DRAFT without sacrificing performance.
        output.write(b"\n  ],\n")
        footer = self.get_footer()
        output.write(orjson.dumps(footer)[1:])
        output.write(b"\n")
        yield output.flush()

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output."""
        if settings.DEBUG:
            return f"/* {exception.__class__.__name__}: {exception} */\n"
        else:
            return f"/* {exception.__class__.__name__} during rendering! */\n"

    def render_feature(
        self, feature_type: FeatureType, instance: models.Model
    ) -> bytes:
        """Render the output of a single feature"""

        # Get all instance attributes:
        properties = self.get_properties(feature_type, feature_type.xsd_type, instance)

        if feature_type.show_name_field:
            name = feature_type.get_display_value(instance)
            geometry_name = b'"geometry_name":%b,' % orjson.dumps(name)
        else:
            geometry_name = b""

        return (
            b"    {"
            b'"type":"Feature",'
            b'"id":%b,'
            b"%b"
            b'"geometry":%b,'
            b'"properties":%b'
            b"}"
        ) % (
            orjson.dumps(f"{feature_type.name}.{instance.pk}"),
            geometry_name,
            self.render_geometry(feature_type, instance),
            orjson.dumps(properties, default=_json_default),
        )

    def _format_geojson_value(self, value):
        if isinstance(value, datetime):
            return value.astimezone(utc)
        elif isinstance(value, models.Model):
            # ForeignKey, not defined as complex type.
            return str(value)
        else:
            return value

    def render_geometry(self, feature_type, instance: models.Model) -> bytes:
        """Generate the proper GeoJSON notation for a geometry.
        This calls the GDAL C-API rendering found in 'GEOSGeometry.json'
        """
        geometry = getattr(instance, feature_type.geometry_field.name)
        if geometry is None:
            return b"null"

        self.output_crs.apply_to(geometry)
        return geometry.json.encode()

    def get_header(self) -> dict:
        """Generate the header fields.

        The format is based on the WFS 3.0 DRAFT. The count fields are moved
        to the footer allowing them to be calculated without performing queries.
        """
        return {
            "type": "FeatureCollection",
            "timeStamp": self._format_geojson_value(self.collection.timestamp),
            # "numberReturned": is written at the end for better query performance.
            "crs": {"type": "name", "properties": {"name": str(self.output_crs)}},
        }

    def get_footer(self) -> dict:
        """Generate the last fields of the response.
        By moving the links, numberReturned/numberMatched fields to the end,
        it's not always needed to perform queries to calculate these.

        The format is based on the WFS 3.0 DRAFT which defines the extra
        pagination headers and numberMatched/numberReturned headers.
        While the draft suggests to put these fields first, there is no such
        requirement in JSON.
        """
        return {
            "links": self.get_links(),
            "numberReturned": self.collection.number_returned,
            "numberMatched": self.collection.number_matched,
        }

    def get_links(self) -> list:
        """Generate the pagination links"""
        links = []
        if self.collection.next:
            links.append(
                {
                    "href": self.collection.next,
                    "rel": "next",
                    "type": "application/geo+json",
                    "title": "next page",
                }
            )
        if self.collection.previous:
            links.append(
                {
                    "href": self.collection.previous,
                    "rel": "previous",
                    "type": "application/geo+json",
                    "title": "previous page",
                }
            )
        return links

    def get_properties(
        self,
        feature_type: FeatureType,
        xsd_type: XsdComplexType,
        instance: models.Model,
    ) -> dict:
        """Collect the data for the 'properties' field.

        This is based on the original XSD definition,
        so the rendering is consistent with other output formats.
        """
        props = {}
        for xsd_element in xsd_type.elements:
            if not xsd_element.is_geometry:
                value = xsd_element.get_value(instance)
                if xsd_element.type.is_complex_type:
                    # Nested object data
                    if value is None:
                        props[xsd_element.name] = None
                    else:
                        value_xsd_type = cast(XsdComplexType, xsd_element.type)
                        if xsd_element.is_many:
                            # If the retrieved QuerySet was not filtered yet, do so now.
                            # This can't be done in get_value() because the FeatureType
                            # is not known there.
                            value = feature_type.filter_related_queryset(value)

                            # "..._to_many relation; reverse FK, M2M or array field.
                            props[xsd_element.name] = [
                                self.get_properties(feature_type, value_xsd_type, item)
                                for item in value
                            ]
                        else:
                            props[xsd_element.name] = self.get_properties(
                                feature_type, value_xsd_type, value
                            )
                else:
                    # Scalar value, or list (for ArrayField).
                    props[xsd_element.name] = self._format_geojson_value(value)

        return props


class DBGeoJsonRenderer(GeoJsonRenderer):
    """GeoJSON renderer that relays the geometry rendering to the database.

    This is even more efficient then calling the C-API for each feature.
    """

    @classmethod
    def decorate_queryset(
        self, feature_type: FeatureType, queryset, output_crs, **params
    ):
        """Update the queryset to let the database render the GML output.
        This is far more efficient then GeoDjango's logic, which performs a
        C-API call for every single coordinate of a geometry.
        """
        queryset = super().decorate_queryset(
            feature_type, queryset, output_crs, **params
        )
        # If desired, the entire FeatureCollection could be rendered
        # in PostgreSQL as well: https://postgis.net/docs/ST_AsGeoJSON.html
        if feature_type.geometry_fields:
            match = feature_type.resolve_element(feature_type.geometry_field.name)
            queryset = queryset.defer(feature_type.geometry_field.name).annotate(
                _as_db_geojson=AsGeoJSON(
                    get_db_geometry_target(match, output_crs),
                    precision=conf.GISSERVER_DB_PRECISION,
                )
            )

        return queryset

    def render_geometry(self, feature_type, instance: models.Model) -> bytes:
        """Generate the proper GeoJSON notation for a geometry"""
        # Database server rendering
        if not feature_type.geometry_fields:
            return b"null"

        geojson = instance._as_db_geojson
        return b"null" if geojson is None else geojson.encode()



from __future__ import annotations

import math

from django.conf import settings
from django.db import models
from django.http import HttpResponse, StreamingHttpResponse
from django.utils.html import escape

from gisserver import conf
from gisserver.exceptions import InvalidParameterValue
from gisserver.features import FeatureRelation, FeatureType
from gisserver.geometries import CRS
from gisserver.operations.base import WFSMethod

from .results import FeatureCollection


class OutputRenderer:
    """Base class to create streaming responses.

    It receives the collected 'context' data of the WFSMethod.
    """

    #: Allow to override the maximum page size.
    #: This value can be 'math.inf' to support endless pages by default.
    max_page_size = None

    #: Define the content type for rendering the output
    content_type = "application/octet-stream"

    #: An optional content-disposition header to output
    content_disposition = None

    def __init__(
        self,
        method: WFSMethod,
        source_query,
        collection: FeatureCollection,
        output_crs: CRS,
    ):
        """
        Receive the collected data to render.

        :param method: The calling WFS Method (e.g. GetFeature class)
        :param source_query: The query that generated this output.
        :type source_query: gisserver.queries.QueryExpression
        :param collection: The collected data for rendering
        :param output_crs: The requested output projection.
        """
        self.method = method
        self.source_query = source_query
        self.collection = collection
        self.output_crs = output_crs
        self.xml_srs_name = escape(str(self.output_crs))

        # Common elements for output rendering:
        self.server_url = method.view.server_url
        self.app_xml_namespace = method.view.xml_namespace

    @classmethod
    def decorate_collection(
        cls, collection: FeatureCollection, output_crs: CRS, **params
    ):
        """Perform presentation-layer logic enhancements on the queryset."""
        for sub_collection in collection.results:
            # Validate the presentation-level parameters for this feature:
            cls.validate(sub_collection.feature_type, **params)

            queryset = cls.decorate_queryset(
                sub_collection.feature_type,
                sub_collection.queryset,
                output_crs,
                **params,
            )
            if queryset is not None:
                sub_collection.queryset = queryset

    @classmethod
    def validate(cls, feature_type: FeatureType, **params):
        """Validate the presentation parameters"""
        crs = params["srsName"]
        if (
            conf.GISSERVER_SUPPORTED_CRS_ONLY
            and crs is not None
            and crs not in feature_type.supported_crs
        ):
            raise InvalidParameterValue(
                "srsName",
                f"Feature '{feature_type.name}' does not support SRID {crs.srid}.",
            )

    @classmethod
    def decorate_queryset(
        cls,
        feature_type: FeatureType,
        queryset: models.QuerySet,
        output_crs: CRS,
        **params,
    ) -> models.QuerySet:
        """Apply presentation layer logic to the queryset."""
        # Avoid fetching relations, fetch these within the same query,
        related = cls._get_prefetch_related(feature_type, output_crs)
        if related:
            queryset = queryset.prefetch_related(*related)

        # Also limit the queryset to the actual fields that are shown.
        # No need to request more data
        fields = [
            f.orm_field
            for f in feature_type.xsd_type.elements
            if not f.is_many or f.is_array  # exclude M2M, but include ArrayField
        ]
        return queryset.only("pk", *fields)

    @classmethod
    def _get_prefetch_related(
        cls, feature_type: FeatureType, output_crs: CRS
    ) -> list[models.Prefetch]:
        """Summarize which fields read data from relations.

        This combines the input from flattened and complex fields,
        in the unlikely case both variations are used in the same feature.
        """
        return [
            models.Prefetch(
                orm_relation.orm_path,
                queryset=cls.get_prefetch_queryset(
                    feature_type, orm_relation, output_crs
                ),
            )
            for orm_relation in feature_type.orm_relations
        ]

    @classmethod
    def get_prefetch_queryset(
        cls,
        feature_type: FeatureType,
        feature_relation: FeatureRelation,
        output_crs: CRS,
    ) -> models.QuerySet | None:
        """Generate a custom queryset that's used to prefetch a relation."""
        # Multiple elements could be referencing the same model, just take first that is filled in.
        if feature_relation.related_model is None:
            return None

        return feature_type.get_related_queryset(feature_relation)

    def get_response(self):
        """Render the output as streaming response."""
        stream = self.render_stream()
        if isinstance(stream, (str, bytes)):
            # Not a real stream, output anyway as regular HTTP response.
            response = HttpResponse(content=stream, content_type=self.content_type)
        else:
            # A actual generator.
            stream = self._trap_exceptions(stream)
            response = StreamingHttpResponse(
                streaming_content=stream,
                content_type=self.content_type,
            )

        # Add HTTP headers
        for name, value in self.get_headers().items():
            response[name] = value

        # Handover to WSGI server (starts streaming when reading the contents)
        return response

    def get_headers(self):
        """Return the response headers"""
        if self.content_disposition:
            # Offer a common quick content-disposition logic that works for all possible queries.
            sub_collection = self.collection.results[0]
            if sub_collection.stop == math.inf:
                if sub_collection.start:
                    page = f"{sub_collection.start}-end"
                else:
                    page = "all"
            elif sub_collection.stop:
                page = f"{sub_collection.start}-{sub_collection.stop - 1}"
            else:
                page = "results"

            return {
                "Content-Disposition": self.content_disposition.format(
                    typenames="+".join(
                        sub.feature_type.name for sub in self.collection.results
                    ),
                    page=page,
                    date=self.collection.date.strftime("%Y-%m-%d %H.%M.%S%z"),
                    timestamp=self.collection.timestamp,
                )
            }

        return {}

    def _trap_exceptions(self, stream):
        """Decorate the generator to show exceptions"""
        try:
            yield from stream
        except Exception as e:
            # Can't return 500 at this point,
            # but can still tell the client what happened.
            yield self.render_exception(e)
            raise

    def render_exception(self, exception: Exception):
        """Inform the client that the stream processing was interrupted with an exception.
        The exception can be rendered in the format fits with the output.

        Purposefully, not much information is given, so avoid informing clients.
        The actual exception is still raised and logged server-side.
        """
        if settings.DEBUG:
            return f"<!-- {exception.__class__.__name__}: {exception} -->\n"
        else:
            return f"<!-- {exception.__class__.__name__} during rendering! -->\n"

    def render_stream(self):
        """Implement this in subclasses to implement a custom output format."""
        raise NotImplementedError()



"""The configuration of a "feature type" in the WFS server.

The "feature type" definitions define what models and attributes are exposed in the WFS server.
When a model attribute is mentioned in the feature type, it can be exposed and queried against.
Any field that is not mentioned in a definition, will therefore not be available, nor queryable.
This metadata is used in the ``GetCapabilities`` call to advertise all available feature types.

To handle other WFS request types besides ``GetCapabilities``, the "feature type" definition
is translated internally into an internal XML Schema Definition (:mod:`gisserver.types`).
That schema maps all model attributes to a specific XML layout, and includes
all XSD Complex Types, elements and attributes linked to the Django model metadata.
The feature type classes (and field types) offer a flexible translation
from attribute listings into a schema definition.
For example, model relationships can be modelled to a different XML layout.
"""
from __future__ import annotations

import html
import operator
from collections import defaultdict
from dataclasses import dataclass
from functools import lru_cache, reduce
from typing import List, Union

from django.conf import settings
from django.contrib.gis.db import models as gis_models
from django.contrib.gis.db.models import Extent, GeometryField
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.db.models.fields.related import ForeignObjectRel  # Django 2.2 import
from django.utils.functional import cached_property  # py3.8: functools

from gisserver.db import conditional_transform
from gisserver.exceptions import ExternalValueError
from gisserver.geometries import CRS, WGS84, BoundingBox
from gisserver.types import (
    GmlBoundedByElement,
    GmlIdAttribute,
    GmlNameElement,
    XPathMatch,
    XsdAnyType,
    XsdComplexType,
    XsdElement,
    XsdTypes,
)

if "django.contrib.postgres" in settings.INSTALLED_APPS:
    from django.contrib.postgres.fields import ArrayField
else:
    ArrayField = None

try:
    from typing import Literal  # Python 3.8

    _all_ = Literal["__all__"]
except ImportError:
    _all_ = str


__all__ = [
    "FeatureType",
    "field",
    "FeatureField",
    "ComplexFeatureField",
    "get_basic_field_type",
]

XSD_TYPES = {
    models.CharField: XsdTypes.string,
    models.TextField: XsdTypes.string,
    models.BooleanField: XsdTypes.boolean,
    models.IntegerField: XsdTypes.integer,
    models.AutoField: XsdTypes.integer,  # Only as of Django 3.0 this extends from IntegerField
    models.FloatField: XsdTypes.double,
    models.DecimalField: XsdTypes.decimal,
    models.TimeField: XsdTypes.time,
    models.DateTimeField: XsdTypes.dateTime,  # note: DateTimeField extends DateField!
    models.DateField: XsdTypes.date,
    models.URLField: XsdTypes.anyURI,
    gis_models.PointField: XsdTypes.gmlPointPropertyType,
    gis_models.PolygonField: XsdTypes.gmlSurfacePropertyType,
    gis_models.LineStringField: XsdTypes.gmlCurvePropertyType,
    gis_models.MultiPointField: XsdTypes.gmlMultiPointPropertyType,
    gis_models.MultiPolygonField: XsdTypes.gmlMultiSurfacePropertyType,
    gis_models.MultiLineStringField: XsdTypes.gmlMultiCurvePropertyType,
    # Generic alternatives
    gis_models.GeometryCollectionField: XsdTypes.gmlMultiGeometryPropertyType,
    gis_models.GeometryField: XsdTypes.gmlGeometryPropertyType,
}
DEFAULT_XSD_TYPE = XsdTypes.anyType


def get_basic_field_type(
    field_name: str, model_field: models.Field | ForeignObjectRel
) -> XsdAnyType:
    """Determine the XSD field type for a Django field."""
    if ArrayField is not None and isinstance(model_field, ArrayField):
        # Determine the type based on the contents.
        # The array notation is written as "is_many"
        model_field = model_field.base_field

    try:
        # Direct instance, quickly resolved!
        return XSD_TYPES[model_field.__class__]
    except KeyError:
        pass

    if isinstance(model_field, models.ForeignKey):
        # Don't let it query on the relation value yet
        return get_basic_field_type(field_name, model_field.target_field)
    elif isinstance(model_field, ForeignObjectRel):
        # e.g. ManyToOneRel descriptor of a foreignkey_id field.
        return get_basic_field_type(field_name, model_field.remote_field.target_field)
    else:
        # Subclass checks:
        for field_cls, xsd_type in XSD_TYPES.items():
            if isinstance(model_field, field_cls):
                return xsd_type

    # Default:
    if isinstance(model_field, GeometryField):
        return XsdTypes.gmlGeometryPropertyType
    else:
        # Default XML choice:
        return DEFAULT_XSD_TYPE


def _get_model_fields(model, fields, parent=None):
    if fields == "__all__":
        # All regular fields
        # Relationships will not be expanded since it can expose so many other fields.
        # Only the relationships that store an ID on the model itself will be exposed.
        return [
            FeatureField(
                name=f.attname,  # using attname so foreignkeys use the name_id field.
                # .bind() is called directly by providing these arguments via init:
                model=model,
                parent=parent,
            )
            for f in model._meta.get_fields()
            if not f.is_relation
            or f.many_to_one  # ForeignKey
            or f.one_to_one  # OneToOneField
        ]
    else:
        # Only defined fields
        fields = [f if isinstance(f, FeatureField) else FeatureField(f) for f in fields]
        for field in fields:
            field.bind(model, parent=parent)
        return fields


@dataclass
class ServiceDescription:
    """Basic metadata for an exposed GIS service."""

    title: str
    abstract: str | None = None
    keywords: list[str] | None = None

    provider_name: str | None = None
    provider_site: str | None = None
    contact_person: str | None = None


class FeatureField:
    """The configuration for a field inside a WFS Feature.

    This defines how a Django model field is mapped into
    an XSD definition that the remaining application uses.
    """

    #: Allow to override the XSD element type that this field will generate.
    xsd_element_class: type[XsdElement] = XsdElement

    model: type[models.Model] | None
    model_field: models.Field | ForeignObjectRel | None

    def __init__(
        self,
        name,
        model_attribute=None,
        model=None,
        parent: ComplexFeatureField | None = None,
        abstract=None,
        xsd_class: type[XsdElement] | None = None,
    ):
        self.name = name
        self.model_attribute = model_attribute
        self.model = None
        self.model_field = None
        self.parent = parent
        self.abstract = abstract

        # Allow to override the class attribute on 'self',
        # which avoids having to subclass this field class as well.
        if xsd_class is not None:
            self.xsd_element_class = xsd_class

        self._nillable_relation = False
        if model is not None:
            self.bind(model)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def _get_xsd_type(self):
        return get_basic_field_type(self.name, self.model_field)

    def bind(
        self,
        model: type[models.Model],
        parent: ComplexFeatureField | None = None,
    ):
        """Late-binding for the model.

        This resolves the model field object from the provided model.
        This method is called internally when the field definition wasn't
        linked to a model yet. This allows the fields to be defined first,
        in external code, and then become part of the ``FeatureType`` fields
        list.

        :param model: The model is field is linked to.
        :param parent: When this element is part of a complex feature,
                       this links to the parent field.
        """
        if self.model is not None:
            raise RuntimeError(f"Feature field '{self.name}' cannot be reused")
        self.model = model
        self.parent = parent

        if self.model_attribute:
            # Support dot based field traversal.
            field_path = self.model_attribute.split(".")
            try:
                field: models.Field = self.model._meta.get_field(field_path[0])
            except FieldDoesNotExist as e:
                raise ImproperlyConfigured(
                    f"FeatureField '{self.name}' has an invalid"
                    f" model_attribute: '{self.model_attribute}' can't be"
                    f" resolved for model '{self.model.__name__}'."
                ) from e

            for name in field_path[1:]:
                if field.null:
                    self._nillable_relation = True

                if not field.is_relation:
                    # Note this tests the previous loop variable, so it checks whether the
                    # field can be walked into in order to resolve the next dotted name below.
                    raise ImproperlyConfigured(
                        f"FeatureField '{field.name}' has an invalid model_attribute: "
                        f"field '{name}' is a '{field.__class__.__name__}', not a relation."
                    )

                field = field.related_model._meta.get_field(name)
            self.model_field = field
        else:
            self.model_field = self.model._meta.get_field(self.name)

    @cached_property
    def xsd_element(self) -> XsdElement:
        """Define the XMLSchema definition for a model field.

        This definition is used by the remaining application to access the
        data. It's the basis for DescribeFeatureType, and it's ``get_value()``
        method is read to access the model field data.
        """
        if self.model_field is None:
            raise RuntimeError(f"bind() was not called for {self!r}")

        # Determine max number of occurrences.
        if isinstance(self.model_field, GeometryField):
            max_occurs = 1  # be explicit here, like mapserver does.
        elif self.model_field.many_to_many or self.model_field.one_to_many:
            max_occurs = "unbounded"  # M2M or reverse FK field
        elif ArrayField is not None and isinstance(self.model_field, ArrayField):
            max_occurs = self.model_field.size or "unbounded"
        else:
            max_occurs = None  # default is 1, but attribute can be left out.

        return self.xsd_element_class(
            name=self.name,
            type=self._get_xsd_type(),
            nillable=self.model_field.null or self._nillable_relation,
            min_occurs=0,
            max_occurs=max_occurs,
            model_attribute=self.model_attribute,
            source=self.model_field,
        )


_FieldDefinition = Union[str, FeatureField]
_FieldDefinitions = Union[_all_, List[_FieldDefinition]]


class ComplexFeatureField(FeatureField):
    """The configuration for an embedded relation field.

    This field type is suitable for any relational object, including
    foreign keys, reverse relations and M2M fields. The internal logic
    translates the relation into an embedded XSD complex type.
    """

    def __init__(
        self,
        name: str,
        fields: _FieldDefinitions,
        model_attribute=None,
        model=None,
        abstract=None,
        xsd_class=None,
        xsd_base_type=XsdTypes.gmlAbstractFeatureType,
    ):
        """
        :param name: Name of the model field.
        :param fields: List of fields to expose for the target model. This can
            be a list of :class:`FeatureField` objects, or plain field names.
            Using ``__all__`` also works but is not recommended outside testing.
        """
        super().__init__(
            name,
            model_attribute=model_attribute,
            model=model,
            abstract=abstract,
            xsd_class=xsd_class,
        )
        self.xsd_base_type = xsd_base_type
        self._fields = fields

    @cached_property
    def fields(self) -> list[FeatureField]:
        """Provide all fields that will be rendered as part of this complex field."""
        return _get_model_fields(self.target_model, self._fields, parent=self)

    def _get_xsd_type(self) -> XsdComplexType:
        """Generate the XSD description for the field with an object relation."""
        pk_field = self.target_model._meta.pk

        return XsdComplexType(
            name=f"{self.target_model._meta.object_name}Type",
            elements=[field.xsd_element for field in self.fields],
            attributes=[
                # Add gml:id attribute definition, so it can be resolved in xpath
                GmlIdAttribute(
                    type_name=self.name,
                    source=pk_field,
                    model_attribute=pk_field.name,
                )
            ],
            base=self.xsd_base_type,
            source=self.target_model,
        )

    @property
    def target_model(self) -> type[models.Model]:
        """Detect which model the relation points to."""
        if self.model_field is None:
            raise RuntimeError("FeatureField.bind() is not called yet")
        elif not self.model_field.is_relation:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} does not support fields of "
                f"type {self.model_field.__class__.__name__}."
            )
        else:
            return self.model_field.related_model


def field(
    name: str,
    *,
    model_attribute=None,
    abstract: str | None = None,
    fields: _FieldDefinitions | None = None,
    xsd_class: type[XsdElement] | None = None,
) -> FeatureField:
    """Shortcut to define a WFS field.

    This automatically selects the proper field class,
    so little knowledge is needed about the internal working.

    :param name: Name of the model field.
    :param fields: If the field exposes a foreign key, provide it's child element names.
        This can be a list of :func:`field` elements, or plain field names.
        Using ``__all__`` also works but is not recommended outside testing.
    """
    if fields is not None:
        return ComplexFeatureField(
            name=name,
            model_attribute=model_attribute,
            fields=fields,
            abstract=abstract,
            xsd_class=xsd_class,
        )
    else:
        return FeatureField(
            name,
            model_attribute=model_attribute,
            abstract=abstract,
            xsd_class=xsd_class,
        )


@dataclass
class FeatureRelation:
    """Tell which related fields are queried by the feature.
    Each dict holds an ORM-path, with the relevant sub-elements.
    """

    #: The ORM path that is queried for this particular relation
    orm_path: str
    #: The fields that will be retrieved for that path
    orm_fields: set[str]
    #: The model that is accessed for this relation (if set)
    related_model: type[models.Model] | None
    #: The source elements that access this relation.
    xsd_elements: list[XsdElement]

    @cached_property
    def _local_model_field_names(self) -> list[str]:
        """Tell which local fields of the model will be accessed by this feature."""

        meta = self.related_model._meta
        result = []
        for name in self.orm_fields:
            model_field = meta.get_field(name)
            if not model_field.many_to_many and not model_field.one_to_many:
                result.append(name)

        # When this relation is retrieved through a ManyToOneRel (reverse FK),
        # the prefetch_related() also needs to have the original foreign key
        # in order to link all prefetches to the proper parent instance.
        for xsd_element in self.xsd_elements:
            if xsd_element.source is not None and xsd_element.source.one_to_many:
                result.append(xsd_element.source.field.name)

        return result


class FeatureType:
    """Declare a feature that is exposed on the map.

    All WFS operations use this class to read the feature ype.
    You may subclass this class to provide extensions,
    such as redefining :meth:`get_queryset`.

    This corresponds with a single Django model.
    """

    #: Allow to override the XSD complex type that this feature will generate.
    xsd_type_class: type[XsdComplexType] = XsdComplexType

    def __init__(
        self,
        queryset: models.QuerySet,
        *,
        fields: _FieldDefinitions | None = None,
        display_field_name: str = None,
        geometry_field_name: str = None,
        name: str = None,
        # WFS Metadata:
        title: str = None,
        abstract: str = None,
        keywords: list[str] = None,
        crs: CRS = None,
        other_crs: list[CRS] = None,
        metadata_url: str | None = None,
        # Settings
        show_name_field: bool = True,
        xml_prefix: str = "app",
    ):
        """
        :param queryset: The queryset to retrieve the data.
        :param fields: Define which fields to show in the WFS data.
            This can be a list of field names, or :class:`FeatureField` objects.
        :param display_field_name: Name of the field that's used as general string representation.
        :param geometry_field: Name of the geometry field to expose (default = auto detect).
        :param name: Name, also used as XML tag name.
        :param title: Used in WFS metadata.
        :param abstract: Used in WFS metadata.
        :param keywords: Used in WFS metadata.
        :param crs: Used in WFS metadata.
        :param other_crs: Used in WFS metadata.
        :param metadata_url: Used in WFS metadata.
        :param show_name_field: Whether to show the ``gml:name`` or the GeoJSON ``geometry_name``
            field. Default is to show a field when ``name_field`` is given.
        :param xml_prefix: The XML namespace prefix to use.
        """
        if isinstance(queryset, models.QuerySet):
            self.queryset = queryset
            self.model = queryset.model
        elif isinstance(queryset, type) and issubclass(queryset, models.Model):
            # In case a model is provided, fix that
            self.model = queryset
            self.queryset = self.model.objects.all()
        else:
            raise TypeError("FeatureType expects a model or queryset")

        self._fields = fields
        self.display_field_name = display_field_name
        self._geometry_field_name = geometry_field_name
        self.name = name or self.model._meta.model_name
        self.title = title or self.model._meta.verbose_name
        self.abstract = abstract
        self.keywords = keywords or []
        self._crs = crs
        self.other_crs = other_crs or []
        self.metadata_url = metadata_url

        # Settings
        self.show_name_field = show_name_field
        self.xml_prefix = xml_prefix

        # Validate that the name doesn't require XML escaping.
        if html.escape(self.name) != self.name or " " in self.name or ":" in self.name:
            raise ValueError(f"Invalid feature name for XML: <{self.xml_name}>")

        self._cached_resolver = lru_cache(100)(self._inner_resolve_element)

    def check_permissions(self, request):
        """Hook that allows subclasses to reject access for datasets.
        It may raise a Django PermissionDenied error.
        """
        pass

    @cached_property
    def xml_name(self):
        """Return the feature name with xml namespace prefix."""
        return f"{self.xml_prefix}:{self.name}"

    @cached_property
    def supported_crs(self) -> list[CRS]:
        """Return all spatial reference system ID's that this feature supports."""
        return [self.crs] + self.other_crs

    @cached_property
    def geometry_fields(self) -> list[GeometryField]:
        """Tell which fields of the model have a geometry field.
        This only compares against fields that are mentioned in this feature type.
        """
        return [
            ff.model_field or getattr(self.model, ff.model_attribute)
            for ff in self.fields
            if ff.xsd_element.is_geometry
        ]

    @cached_property
    def orm_relations(self) -> list[FeatureRelation]:
        """Tell which fields will be retrieved from related fields.

        This gives an object layout based on the XSD elements,
        that can be used for prefetching data.
        """
        models = dict()
        fields = defaultdict(set)
        elements = defaultdict(list)

        # Check all elements that render as "dotted" flattened relation
        for xsd_element in self.xsd_type.flattened_elements:
            if xsd_element.source is not None:
                # Split "relation.field" notation into path, and take the field as child attribute.
                obj_path, field = xsd_element.orm_relation
                elements[obj_path].append(xsd_element)
                fields[obj_path].add(field)
                # field is already on relation:
                models[obj_path] = xsd_element.source.model

        # Check all elements that render as "nested" complex type:
        for xsd_element in self.xsd_type.complex_elements:
            # The complex element itself points to the root of the path,
            # all sub elements become the child attributes.
            obj_path = xsd_element.orm_path
            elements[obj_path].append(xsd_element)
            fields[obj_path] = {
                f.orm_path
                for f in xsd_element.type.elements
                if not f.is_many or f.is_array  # exclude M2M, but include ArrayField
            }
            if xsd_element.source:
                # field references a related object:
                models[obj_path] = xsd_element.source.related_model

        return [
            FeatureRelation(
                orm_path=obj_path,
                orm_fields=sub_fields,
                related_model=models.get(obj_path),
                xsd_elements=elements[obj_path],
            )
            for obj_path, sub_fields in fields.items()
        ]

    @cached_property
    def _local_model_field_names(self) -> list[str]:
        """Tell which local fields of the model will be accessed by this feature."""
        return [
            (ff.model_field.name if ff.model_field else ff.model_attribute)
            for ff in self.fields
            if not ff.model_field.many_to_many and not ff.model_field.one_to_many
        ]

    @cached_property
    def fields(self) -> list[FeatureField]:
        """Define which fields to render."""
        # This lazy reading allows providing 'fields' as lazy value.
        if self._fields is None:
            # Autoconfig, no fields defined.
            if not self._geometry_field_name:
                self._geometry_field_name = next(
                    f.name
                    for f in self.model._meta.get_fields()
                    if isinstance(f, gis_models.GeometryField)
                )

            return [FeatureField(self._geometry_field_name, model=self.model)]
        else:
            return _get_model_fields(self.model, self._fields)

    @cached_property
    def display_field(self) -> models.Field | None:
        """Give access to the field that holds the string-representation."""
        if not self.show_name_field or not self.display_field_name:
            return None
        else:
            return self.model._meta.get_field(self.display_field_name)

    @cached_property
    def geometry_field(self) -> gis_models.GeometryField:
        """Give access to the Django field that holds the geometry."""
        if not self.geometry_fields:
            raise ImproperlyConfigured(
                f"FeatureType '{self.name}' does not expose a geometry field."
            ) from None

        if self._geometry_field_name:
            # Explicitly mentioned, use that field.
            # Check whether the server is not accidentally exposing another field
            # that is not part of the type definition.
            field = next(
                (
                    field
                    for field in self.geometry_fields
                    if field.name == self._geometry_field_name
                ),
                None,
            )
            if field is None:
                raise ImproperlyConfigured(
                    f"FeatureType '{self.name}' does not expose the geometry field "
                    f"'{self._geometry_field_name}' as part of its definition."
                )

            return field
        else:
            # Default: take the first geometry
            return self.geometry_fields[0]

    @cached_property
    def geometry_field_name(self) -> str:
        """Tell which field is the geometry field."""
        # Due to internal reorganization this property is no longer needed,
        # retained for backward compatibility and to reflect any used input parameters.
        return self.geometry_field.name

    @cached_property
    def crs(self) -> CRS:
        """Tell which projection the data should be presented at."""
        if self._crs is None:
            # Default CRS
            return CRS.from_srid(self.geometry_field.srid)  # checks lookup too
        else:
            return self._crs

    def get_queryset(self) -> models.QuerySet:
        """Return the queryset that is used as basis for this feature."""
        # Return a queryset that only retrieves the fields that are actually displayed.
        # That that without .only(), use at least `self.queryset.all()` so a clone is returned.
        return self.queryset.only(*self._local_model_field_names)

    def get_related_queryset(
        self, feature_relation: FeatureRelation
    ) -> models.QuerySet:
        """Return the queryset that is used for prefetching related data."""
        if feature_relation.related_model is None:
            raise RuntimeError(
                f"Unable to create prefetch queryset for relation {feature_relation.orm_path}, "
                f"source model is not defined for: {feature_relation.xsd_elements!r}"
            )
        # Return a queryset that only retrieves the fields that are displayed.
        return self.filter_related_queryset(
            feature_relation.related_model.objects.only(
                *feature_relation._local_model_field_names
            )
        )

    def filter_related_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
        """When a related object returns a queryset, this hook allows extra filtering."""
        return queryset

    def get_bounding_box(self) -> BoundingBox | None:
        """Returns a WGS84 BoundingBox for the complete feature.

        This is used by the GetCapabilities request. It may return ``None``
        when the database table is empty, or the custom queryset doesn't
        return any results.
        """
        if not self.geometry_fields:
            return None

        geo_expression = conditional_transform(
            self.geometry_field.name, self.geometry_field.srid, WGS84.srid
        )

        bbox = self.get_queryset().aggregate(a=Extent(geo_expression))["a"]
        return BoundingBox(*bbox, crs=WGS84) if bbox else None

    def get_envelope(self, instance, crs: CRS | None = None) -> BoundingBox | None:
        """Get the bounding box for a single instance.

        This is only used for native Python rendering. When the database
        rendering is enabled (GISSERVER_USE_DB_RENDERING=True), the calculation
        is entirely performed within the query.
        """
        geometries = [
            geom
            for geom in (getattr(instance, f.name) for f in self.geometry_fields)
            if geom is not None
        ]
        if not geometries:
            return None

        # Perform the combining of geometries inside libgeos
        geometry = (
            geometries[0] if len(geometries) == 1 else reduce(operator.or_, geometries)
        )
        if crs is not None and geometry.srid != crs.srid:
            crs.apply_to(geometry)  # avoid clone
        return BoundingBox.from_geometry(geometry, crs=crs)

    def get_display_value(self, instance: models.Model) -> str:
        """Generate the display name value"""
        if self.display_field_name:
            return getattr(instance, self.display_field_name)
        else:
            return str(instance)

    @property
    def xsd_base_type(self) -> XsdAnyType:
        """Return the base class for the :attr:`xsd_type`.

        This builds an XsdComplexType element that represents
        the contents of :attr:`XsdTypes.gmlAbstractFeatureType`.
        By providing this as Complex Type, the filters can also resolve the
        attributes of ``@gml:id``, ``<gml:name>`` and ``<gml:boundedBy>`` nodes.
        """
        pk_field = self.model._meta.pk

        # Define <gml:boundedBy>, if the feature has a geometry
        base_elements = []
        if self.geometry_fields:
            # Without a geometry, boundaries are not possible.
            base_elements.append(GmlBoundedByElement(feature_type=self))

        if self.show_name_field:
            # Add <gml:name>
            gml_name = GmlNameElement(
                model_attribute=self.display_field_name,
                source=self.display_field,
                feature_type=self,
            )
            base_elements.insert(0, gml_name)

        return XsdComplexType(
            prefix="gml",
            name="AbstractFeatureType",
            elements=base_elements,
            attributes=[
                # Add gml:id attribute definition so it can be resolved in xpath
                GmlIdAttribute(
                    type_name=self.name,
                    source=pk_field,
                    model_attribute=pk_field.name,
                )
            ],
            base=XsdTypes.gmlAbstractGMLType,
        )

    @cached_property
    def xsd_type(self) -> XsdComplexType:
        """Return the definition of this feature as an XSD Complex Type."""
        return self.xsd_type_class(
            name=f"{self.name.title()}Type",
            elements=[field.xsd_element for field in self.fields],
            base=self.xsd_base_type,
            source=self.model,
        )

    def resolve_element(self, xpath: str) -> XPathMatch:
        """Resolve the element, and the matching object.

        This is used to convert XPath references in requests
        to the actual elements and model attributes for queries.

        Internally, this method caches results.
        """
        nodes = self._cached_resolver(xpath)  # calls _inner_resolve_element
        if nodes is None:
            raise ExternalValueError(f"Field '{xpath}' does not exist.")

        return XPathMatch(self, nodes, query=xpath)

    def _inner_resolve_element(self, xpath: str):
        """Inner part of resolve_element() that is cached.
        This performs any additional checks that happen at the root-level only.
        """
        # Avoid complex XPath features for now. Only element/child works.
        # The attribute selector is not fully implemented, but tested elsewhere.
        if "//" in xpath:
            raise NotImplementedError(
                f"XPath selectors with deeper descendant selectors are not supported: {xpath}"
            )
        elif "::" in xpath:
            raise NotImplementedError(
                f"XPath selectors with expanded syntax are not supported: {xpath}"
            )
        elif "(" in xpath:
            raise NotImplementedError(
                f"XPath selectors with functions are not supported: {xpath}"
            )

        # Allow /app:ElementName/.. as "absolute" path.
        # Given our internal resolver logic, simple solution is to strip it.
        for root_prefix in (
            f"{self.name}/",
            f"/{self.name}/",
            f"{self.xml_name}/",
            f"/{self.xml_name}/",
        ):
            if xpath.startswith(root_prefix):
                xpath = xpath[len(root_prefix) :]
                break

        return self.xsd_type.resolve_element_path(xpath)



"""Handle adhoc-query objects.

The adhoc query is based on incoming request parameters,
such as the "FILTER", "BBOX" and "RESOURCEID" parameters.

These definitions follow the WFS spec.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass

from django.db.models import Q

from gisserver import conf
from gisserver.exceptions import InvalidParameterValue, MissingParameterValue
from gisserver.features import FeatureType
from gisserver.geometries import BoundingBox
from gisserver.parsers import fes20
from gisserver.parsers.fes20 import operators

from .base import QueryExpression

logger = logging.getLogger(__name__)


@dataclass
class AdhocQuery(QueryExpression):
    """The Ad hoc query expression parameters.

    This represents all dynamic queries received as request (hence "adhoc"),
    such as the "FILTER" and "BBOX" arguments from a HTTP GET.

    The WFS Spec has 3 class levels for this:
    - AdhocQueryExpression (types, projection, selection, sorting)
    - Query (adds srsName, featureVersion)

    For KVP requests, this dataclass is almost identical to **params.
    However, it allows combining the filter parameters. These become
    one single XML request for HTTP POST requests later.
    """

    typeNames: list[FeatureType]  # typeNames in WFS/FES spec
    # aliases: Optional[List[str]] = None
    handle: str = ""  # only for XML POST requests

    # Projection clause:
    # propertyName

    # Selection clause:
    # - for XML POST this is encoded in a <fes:Query>
    # - for HTTP GET, this is encoded as FILTER, FILTER_LANGUAGE, RESOURCEID, BBOX.
    filter: fes20.Filter | None = None
    filter_language: str = fes20.Filter.query_language
    bbox: BoundingBox | None = None

    # Sorting Clause
    sortBy: fes20.SortBy | None = None

    # Officially part of the GetFeature/GetPropertyValue request object,
    # but included here for ease of query implementation.
    resourceId: fes20.IdOperator | None = None

    # GetPropertyValue:
    # In the WFS spec, this is only part of the operation/presentation.
    # For Django, we'd like to make this part of the query too.
    value_reference: fes20.ValueReference | None = None

    @classmethod
    def from_kvp_request(cls, **params):
        """Build this object from a HTTP GET (key-value-pair) request."""
        # Validate optionally required parameters
        if not params["typeNames"] and not params["resourceID"]:
            raise MissingParameterValue("typeNames", "Empty TYPENAMES parameter")

        # Validate mutually exclusive parameters
        if params["filter"] and (params["bbox"] or params["resourceID"]):
            raise InvalidParameterValue(
                "filter",
                "The FILTER parameter is mutually exclusive with BBOX and RESOURCEID",
            )

        # Validate mutually exclusive parameters
        if params["resourceID"]:
            if params["bbox"] or params["filter"]:
                raise InvalidParameterValue(
                    "resourceID",
                    "The RESOURCEID parameter is mutually exclusive with BBOX and FILTER",
                )

            # When ResourceId + typenames is defined, it should be a value from typenames
            # see WFS spec 7.9.2.4.1
            if params["typeNames"]:
                id_type_names = params["resourceID"].type_names
                if id_type_names:
                    # Only test when the RESOURCEID has a typename.id format
                    # Otherwise, this breaks the CITE RESOURCEID=test-UUID parameter.
                    kvp_type_names = {
                        feature_type.name for feature_type in params["typeNames"]
                    }
                    if not kvp_type_names.issuperset(id_type_names):
                        raise InvalidParameterValue(
                            "resourceID",
                            "When TYPENAMES and RESOURCEID are combined, "
                            "the RESOURCEID type should be included in TYPENAMES.",
                        )

        return AdhocQuery(
            typeNames=params["typeNames"],
            filter=params["filter"],
            filter_language=params["filter_language"],
            bbox=params["bbox"],
            sortBy=params["sortBy"],
            resourceId=params["resourceID"],
            value_reference=params.get("valueReference"),
        )

    def bind(self, *args, **kwargs):
        """Inform this quey object of the available feature types"""
        super().bind(*args, **kwargs)

        if self.resourceId:
            # Early validation whether the selected resourceID type exists.
            feature_types = [
                self.resolve_type_name(type_name, locator="resourceID")
                for type_name in self.resourceId.type_names
            ]

            # Also make the behavior consistent, always supply the type name.
            if not self.typeNames:
                self.typeNames = feature_types

    def get_type_names(self):
        return self.typeNames

    def compile_query(
        self, feature_type: FeatureType, using=None
    ) -> fes20.CompiledQuery:
        """Return our internal CompiledQuery object that can be applied to the queryset."""
        if self.filter:
            # Generate the internal query object from the <fes:Filter>
            return self.filter.compile_query(feature_type, using=using)
        else:
            # Generate the internal query object from the BBOX and sortBy args.
            return self._compile_non_filter_query(feature_type, using=using)

    def _compile_non_filter_query(self, feature_type: FeatureType, using=None):
        """Generate the query based on the remaining parameters.

        This is slightly more efficient then generating the fes Filter object
        from these KVP parameters (which could also be done within the request method).
        """
        compiler = fes20.CompiledQuery(feature_type=feature_type, using=using)

        if self.bbox:
            # Validate whether the provided SRID is supported.
            # While PostGIS would support many more ID's,
            # it would crash when an unsupported ID is given.
            crs = self.bbox.crs
            if (
                conf.GISSERVER_SUPPORTED_CRS_ONLY
                and crs is not None
                and crs not in feature_type.supported_crs
            ):
                raise InvalidParameterValue(
                    "bbox",
                    f"Feature '{feature_type.name}' does not support SRID {crs.srid}.",
                )

            # Using __within does not work with geometries
            # that only partially exist within the bbox
            lookup = operators.SpatialOperatorName.BBOX.value  # "intersects"
            filters = {
                f"{feature_type.geometry_field.name}__{lookup}": self.bbox.as_polygon(),
            }
            compiler.add_lookups(Q(**filters))

        if self.resourceId:
            self.resourceId.build_query(compiler=compiler)

        if self.sortBy:
            compiler.add_sort_by(self.sortBy)

        return compiler



"""All parser logic to process <fes:...> and <gml:...> tags."""

from .fes20 import Filter
from .fes20 import function_registry as fes_function_registry
from .gml import parse_gml

parse_fes = Filter.from_string

__all__ = [
    "parse_fes",
    "parse_gml",
    "fes_function_registry",
]



"""Internal module for additional GIS database functions."""
from __future__ import annotations

from functools import reduce
from typing import cast

from django.contrib.gis.db.models import GeometryField, functions
from django.db import connection, connections, models

from gisserver import conf
from gisserver.geometries import CRS
from gisserver.types import XPathMatch, XsdElement


class AsEWKT(functions.GeoFunc):
    """Generate EWKT in the database (PostGIS tested only at the moment)."""

    name = "AsEWKT"
    output_field = models.TextField()

    def __init__(self, field, precision=conf.GISSERVER_DB_PRECISION, **extra):
        if connection.ops.spatial_version >= (3, 1):
            super().__init__(field, precision, **extra)
        else:
            super().__init__(field, **extra)


class AsGML(functions.AsGML):
    name = "AsGML"

    def __init__(
        self,
        expression,
        version=3,
        precision=conf.GISSERVER_DB_PRECISION,
        envelope=False,
        **extra,
    ):
        # Note that Django's AsGml, version=2, precision=8
        # the options is postgres-only.
        super().__init__(expression, version, precision, **extra)
        self.envelope = envelope

    def as_postgresql(self, compiler, connection, **extra_context):
        # Fill options parameter (https://postgis.net/docs/ST_AsGML.html)
        options = 33 if self.envelope else 1  # 32 = bbox, 1 = long CRS urn
        template = f"%(function)s(%(expressions)s, {options})"
        return self.as_sql(compiler, connection, template=template, **extra_context)


class ST_Union(functions.Union):
    name = "Union"
    arity = None

    def as_postgresql(self, compiler, connection, **extra_context):
        # PostgreSQL can handle ST_Union(ARRAY(field names)), other databases don't.
        if len(self.source_expressions) > 2:
            extra_context["template"] = "%(function)s(ARRAY(%(expressions)s))"
        return self.as_sql(compiler, connection, **extra_context)


def get_geometries_union(
    expressions: list[str | functions.GeoFunc], using="default"
) -> str | functions.Union:
    """Generate a union of multiple geometry fields."""
    if not expressions:
        raise ValueError("Missing geometery fields for get_geometries_union()")

    if len(expressions) == 1:
        return next(iter(expressions))  # fastest in set data type
    elif len(expressions) == 2:
        return functions.Union(*expressions)
    elif connections[using].vendor == "postgresql":
        # postgres can handle multiple field names
        return ST_Union(expressions)
    else:
        # other databases do Union(Union(1, 2), 3)
        return reduce(functions.Union, expressions)


def conditional_transform(
    expression: str | functions.GeoFunc, expression_srid: int, output_srid: int
) -> str | functions.Transform:
    """Apply a CRS Transform to the queried field if that is needed."""
    if expression_srid != output_srid:
        expression = functions.Transform(expression, srid=output_srid)
    return expression


def build_db_annotations(
    selects: dict[str, str | functions.Func],
    name_template: str,
    wrapper_func: type[functions.Func],
) -> dict:
    """Utility to build annotations for all geometry fields for an XSD type.
    This is used by various DB-optimized rendering methods.
    """
    return {
        escape_xml_name(name, name_template): wrapper_func(target)
        for name, target in selects.items()
    }


def get_db_annotation(instance: models.Model, name: str, name_template: str):
    """Retrieve the value that an annotation has added to the model."""
    # The "name" allows any XML-tag elements, escape the most obvious
    escaped_name = escape_xml_name(name, name_template)
    try:
        return getattr(instance, escaped_name)
    except AttributeError as e:
        raise AttributeError(
            f" DB annotation {instance._meta.model_name}.{escaped_name}"
            f" not found (using {name_template})"
        ) from e


def escape_xml_name(name: str, template="{name}") -> str:
    """Escape an XML name to be used as annotation name."""
    return template.format(name=name.replace(".", "_"))


def get_db_geometry_selects(
    gml_elements: list[XsdElement], output_crs: CRS
) -> dict[str, str | functions.Transform]:
    """Utility to generate select clauses for the geometry fields of a type.
    Key is the xsd element name, value is the database select expression.
    """
    return {
        xsd_element.name: _get_db_geometry_target(xsd_element, output_crs)
        for xsd_element in gml_elements
        if xsd_element.source is not None
    }


def _get_db_geometry_target(
    xsd_element: XsdElement, output_crs: CRS
) -> str | functions.Transform:
    """Wrap the selection of a geometry field in a CRS Transform if needed."""
    field = cast(GeometryField, xsd_element.source)
    return conditional_transform(
        xsd_element.orm_path, field.srid, output_srid=output_crs.srid
    )


def get_db_geometry_target(xpath_match: XPathMatch, output_crs: CRS):
    """Based on a resolved element, build the proper geometry field select clause."""
    field = cast(GeometryField, xpath_match.child.source)
    return conditional_transform(
        xpath_match.orm_path, field.srid, output_srid=output_crs.srid
    )
